from __future__ import annotations

import asyncio
import json
import logging
import sys
from typing import TYPE_CHECKING

import click

from ai.backend.cli.types import ExitCode
from ai.backend.common.cli import EnumChoice, MinMaxRange
from ai.backend.common.etcd import ConfigScopes
from ai.backend.common.etcd import quote as etcd_quote
from ai.backend.common.etcd import unquote as etcd_unquote
from ai.backend.common.logging import BraceStyleAdapter

from .context import etcd_ctx
from .image_impl import alias as alias_impl
from .image_impl import dealias as dealias_impl
from .image_impl import forget_image as forget_image_impl
from .image_impl import inspect_image as inspect_image_impl
from .image_impl import list_images as list_images_impl
from .image_impl import rescan_images as rescan_images_impl
from .image_impl import set_image_resource_limit as set_image_resource_limit_impl

if TYPE_CHECKING:
    from .context import CLIContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.argument("key")
@click.argument("value")
@click.option(
    "-s",
    "--scope",
    type=EnumChoice(ConfigScopes),
    default=ConfigScopes.GLOBAL,
    help="The configuration scope to put the value.",
)
@click.pass_obj
def put(cli_ctx: CLIContext, key, value, scope) -> None:
    """Put a single key-value pair into the etcd."""

    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                await etcd.put(key, value, scope=scope)
            except Exception:
                log.exception("An error occurred.")

    asyncio.run(_impl())


@cli.command()
@click.argument("key", type=str)
@click.argument("file", type=click.File("rb"))
@click.option(
    "-s",
    "--scope",
    type=EnumChoice(ConfigScopes),
    default=ConfigScopes.GLOBAL,
    help="The configuration scope to put the value.",
)
@click.pass_obj
def put_json(cli_ctx: CLIContext, key, file, scope) -> None:
    """
    Put a JSON object from FILE to the etcd as flattened key-value pairs
    under the given KEY prefix.
    """

    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                value = json.load(file)
                await etcd.put_prefix(key, value, scope=scope)
            except Exception:
                log.exception("An error occurred.")

    asyncio.run(_impl())


@cli.command()
@click.argument("src_prefix", type=str)
@click.argument("dst_prefix", type=str)
@click.option(
    "-s",
    "--scope",
    type=EnumChoice(ConfigScopes),
    default=ConfigScopes.GLOBAL,
    help=(
        "The configuration scope to get/put the subtree. "
        "To move between different scopes, use the global scope "
        "and specify the per-scope prefixes manually."
    ),
)
@click.pass_obj
def move_subtree(cli_ctx: CLIContext, src_prefix, dst_prefix, scope) -> None:
    """
    Move a subtree to another key prefix.
    """

    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                subtree = await etcd.get_prefix(src_prefix, scope=scope)
                await etcd.put_prefix(dst_prefix, subtree, scope=scope)
                await etcd.delete_prefix(src_prefix, scope=scope)
            except Exception:
                log.exception("An error occurred.")

    asyncio.run(_impl())


@cli.command()
@click.argument("key")
@click.option(
    "--prefix",
    is_flag=True,
    help="Get all key-value pairs prefixed with the given key as a JSON form.",
)
@click.option(
    "-s",
    "--scope",
    type=EnumChoice(ConfigScopes),
    default=ConfigScopes.GLOBAL,
    help="The configuration scope to put the value.",
)
@click.pass_obj
def get(cli_ctx: CLIContext, key, prefix, scope) -> None:
    """
    Get the value of a key in the configured etcd namespace.
    """

    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                if prefix:
                    data = await etcd.get_prefix(key, scope=scope)
                    print(json.dumps(dict(data), indent=4))
                else:
                    val = await etcd.get(key, scope=scope)
                    if val is None:
                        sys.exit(ExitCode.FAILURE)
                    print(val)
            except Exception:
                log.exception("An error occurred.")

    asyncio.run(_impl())


@cli.command()
@click.argument("key")
@click.option("--prefix", is_flag=True, help="Delete all keys prefixed with the given key.")
@click.option(
    "-s",
    "--scope",
    type=EnumChoice(ConfigScopes),
    default=ConfigScopes.GLOBAL,
    help="The configuration scope to put the value.",
)
@click.pass_obj
def delete(cli_ctx: CLIContext, key, prefix, scope) -> None:
    """Delete the key in the configured etcd namespace."""

    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                if prefix:
                    data = await etcd.get_prefix(key, scope=scope)
                    if not data:
                        log.info(f"No keys found to delete with prefix: {key}")
                        return
                    await etcd.delete_prefix(key, scope=scope)
                    log.info(f"All keys starting with '{key}' successfully deleted.")
                else:
                    data = await etcd.get(key, scope=scope)
                    if data is None:
                        log.info(f"No key found to delete: {key}")
                        return
                    await etcd.delete(key, scope=scope)
                log.info(f"Key '{key}' successfully deleted.")
            except Exception:
                log.exception("An error occurred.")

    asyncio.run(_impl())


@cli.command()
@click.option("-s", "--short", is_flag=True, help="Show only the image references and digests.")
@click.option("-i", "--installed", is_flag=True, help="Show only the installed images.")
@click.pass_obj
def list_images(cli_ctx, short, installed) -> None:
    """List all configured images."""
    log.warning("etcd list-images command is deprecated, use image list instead")
    asyncio.run(list_images_impl(cli_ctx, short, installed))


@cli.command()
@click.argument("canonical_or_alias")
@click.argument("architecture")
@click.pass_obj
def inspect_image(cli_ctx, canonical_or_alias, architecture) -> None:
    """Show the details of the given image or alias."""
    log.warning("etcd inspect-image command is deprecated, use image inspect instead")
    asyncio.run(inspect_image_impl(cli_ctx, canonical_or_alias, architecture))


@cli.command()
@click.argument("canonical_or_alias")
@click.argument("architecture")
@click.pass_obj
def forget_image(cli_ctx, canonical_or_alias, architecture) -> None:
    """Forget (delete) a specific image."""
    log.warning("etcd forget-image command is deprecated, use image forget instead")
    asyncio.run(forget_image_impl(cli_ctx, canonical_or_alias, architecture))


@cli.command()
@click.argument("canonical_or_alias")
@click.argument("slot_type")
@click.argument("range_value", type=MinMaxRange)
@click.argument("architecture")
@click.pass_obj
def set_image_resource_limit(
    cli_ctx,
    canonical_or_alias,
    slot_type,
    range_value,
    architecture,
) -> None:
    """Set the MIN:MAX values of a SLOT_TYPE limit for the given image REFERENCE."""
    log.warning(
        "etcd set-image-resource-limit command is deprecated, use image set-resource-limit instead"
    )
    asyncio.run(
        set_image_resource_limit_impl(
            cli_ctx,
            canonical_or_alias,
            slot_type,
            range_value,
            architecture,
        )
    )


@cli.command()
@click.argument("registry")
@click.pass_obj
def rescan_images(cli_ctx: CLIContext, registry: str) -> None:
    """
    Update the kernel image metadata from all configured docker registries.

    Pass the name (usually hostname or "lablup") of the Docker registry configured as REGISTRY.
    """
    log.warning("etcd rescan-images command is deprecated, use image rescan instead")
    asyncio.run(rescan_images_impl(cli_ctx, registry, False))


@cli.command()
@click.argument("alias")
@click.argument("target")
@click.argument("architecture")
@click.pass_obj
def alias(cli_ctx: CLIContext, alias: str, target: str, architecture: str) -> None:
    """Add an image alias from the given alias to the target image reference."""
    log.warning("etcd alias command is deprecated, use image alias instead")
    asyncio.run(alias_impl(cli_ctx, alias, target, architecture))


@cli.command()
@click.argument("alias")
@click.pass_obj
def dealias(cli_ctx: CLIContext, alias: str) -> None:
    """Remove an alias."""
    log.warning("etcd dealias command is deprecated, use image dealias instead")
    asyncio.run(dealias_impl(cli_ctx, alias))


@cli.command()
@click.argument("value")
@click.pass_obj
def quote(cli_ctx: CLIContext, value: str) -> None:
    """
    Quote the given string for use as a URL piece in etcd keys.
    Use this to generate argument inputs for aliases and raw image keys.
    """
    print(etcd_quote(value))


@cli.command()
@click.argument("value")
@click.pass_obj
def unquote(cli_ctx: CLIContext, value: str) -> None:
    """
    Unquote the given string used as a URL piece in etcd keys.
    """
    print(etcd_unquote(value))


@cli.command()
@click.argument("proxy")
@click.argument("scaling_groups")
@click.option(
    "-s",
    "--scope",
    type=EnumChoice(ConfigScopes),
    default=ConfigScopes.GLOBAL,
    help="The configuration scope to put the value.",
)
@click.pass_obj
def set_storage_sftp_scaling_group(
    cli_ctx: CLIContext,
    proxy: str,
    scaling_groups: str,
    scope: ConfigScopes,
) -> None:
    """
    Updates storage proxy node config's SFTP desginated scaling groups.
    To enter multiple scaling groups concatenate names with comma(,).
    """

    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            data = await etcd.get_prefix(f"volumes/proxies/{proxy}", scope=scope)
            if len(data) == 0:
                log.error("proxy {} does not exist", proxy)
                sys.exit(ExitCode.FAILURE)
            await etcd.put(
                f"volumes/proxies/{proxy}/sftp_scaling_groups",
                ",".join([x.strip() for x in scaling_groups.split(",")]),
            )

    asyncio.run(_impl())


@cli.command()
@click.argument("proxy")
@click.option(
    "-s",
    "--scope",
    type=EnumChoice(ConfigScopes),
    default=ConfigScopes.GLOBAL,
    help="The configuration scope to put the value.",
)
@click.pass_obj
def remove_storage_sftp_scaling_group(
    cli_ctx: CLIContext,
    proxy: str,
    scope: ConfigScopes,
) -> None:
    """
    Removes storage proxy node config's SFTP desginated scaling groups.
    """

    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            data = await etcd.get_prefix(f"volumes/proxies/{proxy}", scope=scope)
            if len(data) == 0:
                log.error("proxy {} does not exist", proxy)
                sys.exit(ExitCode.FAILURE)
            await etcd.delete(f"volumes/proxies/{proxy}/sftp_scaling_groups")

    asyncio.run(_impl())



from __future__ import annotations

import base64
import secrets
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Sequence, Tuple, TypedDict

import graphene
import sqlalchemy as sa
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from dateutil.parser import parse as dtparse
from graphene.types.datetime import DateTime as GQLDateTime
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.orm import relationship
from sqlalchemy.sql.expression import false

from ai.backend.common import msgpack, redis_helper
from ai.backend.common.types import AccessKey, SecretKey

if TYPE_CHECKING:
    from .gql import GraphQueryContext
    from .vfolder import VirtualFolder

from ..defs import RESERVED_DOTFILES
from .base import (
    Base,
    ForeignKeyIDColumn,
    Item,
    PaginatedList,
    batch_multiresult,
    batch_result,
    mapper_registry,
    set_if_set,
    simple_db_mutate,
    simple_db_mutate_returning_item,
)
from .minilang.ordering import OrderSpecItem, QueryOrderParser
from .minilang.queryfilter import FieldSpecItem, QueryFilterParser
from .user import ModifyUserInput, UserRole
from .utils import agg_to_array

__all__: Sequence[str] = (
    "keypairs",
    "KeyPairRow",
    "KeyPair",
    "KeyPairList",
    "UserInfo",
    "KeyPairInput",
    "CreateKeyPair",
    "ModifyKeyPair",
    "DeleteKeyPair",
    "Dotfile",
    "MAXIMUM_DOTFILE_SIZE",
    "query_owned_dotfiles",
    "query_bootstrap_script",
    "verify_dotfile_name",
)


MAXIMUM_DOTFILE_SIZE = 64 * 1024  # 61 KiB

keypairs = sa.Table(
    "keypairs",
    mapper_registry.metadata,
    sa.Column("user_id", sa.String(length=256), index=True),
    sa.Column("access_key", sa.String(length=20), primary_key=True),
    sa.Column("secret_key", sa.String(length=40)),
    sa.Column("is_active", sa.Boolean, index=True),
    sa.Column("is_admin", sa.Boolean, index=True, default=False, server_default=false()),
    sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    sa.Column(
        "modified_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        onupdate=sa.func.current_timestamp(),
    ),
    sa.Column("last_used", sa.DateTime(timezone=True), nullable=True),
    sa.Column("rate_limit", sa.Integer),
    sa.Column("num_queries", sa.Integer, server_default="0"),
    # SSH Keypairs.
    sa.Column("ssh_public_key", sa.Text, nullable=True),
    sa.Column("ssh_private_key", sa.Text, nullable=True),
    ForeignKeyIDColumn("user", "users.uuid", nullable=False),
    sa.Column(
        "resource_policy",
        sa.String(length=256),
        sa.ForeignKey("keypair_resource_policies.name"),
        nullable=False,
    ),
    # dotfiles column, \x90 means empty list in msgpack
    sa.Column(
        "dotfiles", sa.LargeBinary(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=b"\x90"
    ),
    sa.Column(
        "bootstrap_script", sa.String(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=""
    ),
)


class KeyPairRow(Base):
    __table__ = keypairs
    sessions = relationship("SessionRow", back_populates="access_key_row")
    resource_policy_row = relationship("KeyPairResourcePolicyRow", back_populates="keypairs")
    scaling_groups = relationship(
        "ScalingGroupRow",
        secondary="sgroups_for_keypairs",
        back_populates="keypairs",
    )

    user_row = relationship("UserRow", back_populates="keypairs", foreign_keys=keypairs.c.user)


class UserInfo(graphene.ObjectType):
    email = graphene.String()
    full_name = graphene.String()

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row,
    ) -> Optional[UserInfo]:
        if row is None:
            return None
        return cls(email=row["email"], full_name=row["full_name"])

    @classmethod
    async def batch_load_by_uuid(
        cls,
        ctx: GraphQueryContext,
        user_uuids: Sequence[uuid.UUID],
    ) -> Sequence[Optional[UserInfo]]:
        async with ctx.db.begin_readonly() as conn:
            from .user import users

            query = (
                sa.select([users.c.uuid, users.c.email, users.c.full_name])
                .select_from(users)
                .where(users.c.uuid.in_(user_uuids))
            )
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                user_uuids,
                lambda row: row["uuid"],
            )


class KeyPair(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    user_id = graphene.String()
    full_name = graphene.String()
    access_key = graphene.String()
    secret_key = graphene.String()
    is_active = graphene.Boolean()
    is_admin = graphene.Boolean()
    resource_policy = graphene.String()
    created_at = GQLDateTime()
    last_used = GQLDateTime()
    rate_limit = graphene.Int()
    num_queries = graphene.Int()
    user = graphene.UUID()
    projects = graphene.List(lambda: graphene.String)

    ssh_public_key = graphene.String()

    vfolders = graphene.List("ai.backend.manager.models.VirtualFolder")
    compute_sessions = graphene.List(
        "ai.backend.manager.models.ComputeSession",
        status=graphene.String(),
    )
    concurrency_used = graphene.Int()

    user_info = graphene.Field(lambda: UserInfo)

    # Deprecated
    concurrency_limit = graphene.Int(
        deprecation_reason=(
            "Moved to KeyPairResourcePolicy object as the max_concurrent_sessions field."
        )
    )

    async def resolve_user_info(
        self,
        info: graphene.ResolveInfo,
    ) -> UserInfo:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader(ctx, "UserInfo.by_uuid")
        return await loader.load(self.user)

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row,
    ) -> KeyPair:
        return cls(
            id=row["access_key"],
            user_id=row["user_id"],
            full_name=row["full_name"] if "full_name" in row.keys() else None,
            access_key=row["access_key"],
            secret_key=row["secret_key"],
            is_active=row["is_active"],
            is_admin=row["is_admin"],
            resource_policy=row["resource_policy"],
            created_at=row["created_at"],
            rate_limit=row["rate_limit"],
            user=row["user"],
            ssh_public_key=row["ssh_public_key"],
            concurrency_limit=0,  # deprecated
            projects=row["groups_name"] if "groups_name" in row.keys() else [],
        )

    async def resolve_num_queries(self, info: graphene.ResolveInfo) -> int:
        ctx: GraphQueryContext = info.context
        n = await redis_helper.execute(
            ctx.redis_stat, lambda r: r.get(f"kp:{self.access_key}:num_queries")
        )
        if n is not None:
            return n
        return 0

    async def resolve_vfolders(self, info: graphene.ResolveInfo) -> Sequence[VirtualFolder]:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader(ctx, "VirtualFolder")
        return await loader.load(self.access_key)

    async def resolve_compute_sessions(
        self, info: graphene.ResolveInfo, raw_status: Optional[str] = None
    ):
        ctx: GraphQueryContext = info.context
        from . import KernelStatus  # noqa: avoid circular imports

        if raw_status is not None:
            status = KernelStatus[raw_status]
        loader = ctx.dataloader_manager.get_loader(ctx, "ComputeSession", status=status)
        return await loader.load(self.access_key)

    async def resolve_concurrency_used(self, info: graphene.ResolveInfo) -> int:
        ctx: GraphQueryContext = info.context
        kp_key = "keypair.concurrency_used"
        concurrency_used = await redis_helper.execute(
            ctx.redis_stat,
            lambda r: r.get(f"{kp_key}.{self.access_key}"),
        )
        if concurrency_used is not None:
            return int(concurrency_used)
        return 0

    async def resolve_last_used(self, info: graphene.ResolveInfo) -> datetime | None:
        ctx: GraphQueryContext = info.context
        last_call_time_key = f"kp:{self.access_key}:last_call_time"
        row_ts = await redis_helper.execute(ctx.redis_stat, lambda r: r.get(last_call_time_key))
        if row_ts is None:
            return None
        return datetime.fromtimestamp(float(row_ts))

    @classmethod
    async def load_all(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        domain_name: Optional[str] = None,
        is_active: Optional[bool] = None,
        limit: Optional[int] = None,
    ) -> Sequence[KeyPair]:
        from .user import users

        j = sa.join(
            keypairs,
            users,
            keypairs.c.user == users.c.uuid,
        )
        query = sa.select([keypairs]).select_from(j)
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        if limit is not None:
            query = query.limit(limit)
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(graph_ctx, row)) is not None
            ]

    _queryfilter_fieldspec: Mapping[str, FieldSpecItem] = {
        "access_key": ("keypairs_access_key", None),
        "user_id": ("users_uuid", None),
        "email": ("users_email", None),
        "full_name": ("users_full_name", None),
        "is_active": ("keypairs_is_active", None),
        "is_admin": ("keypairs_is_admin", None),
        "resource_policy": ("keypairs_resource_policy", None),
        "created_at": ("keypairs_created_at", dtparse),
        "last_used": ("keypairs_last_used", dtparse),
        "rate_limit": ("keypairs_rate_limit", None),
        "num_queries": ("keypairs_num_queries", None),
        "ssh_public_key": ("keypairs_ssh_public_key", None),
        "projects": ("groups_name", None),
    }

    _queryorder_colmap: Mapping[str, OrderSpecItem] = {
        "access_key": ("keypairs_access_key", None),
        "email": ("users_email", None),
        "full_name": ("users_full_name", None),
        "is_active": ("keypairs_is_active", None),
        "is_admin": ("keypairs_is_admin", None),
        "resource_policy": ("keypairs_resource_policy", None),
        "created_at": ("keypairs_created_at", None),
        "last_used": ("keypairs_last_used", None),
        "rate_limit": ("keypairs_rate_limit", None),
        "num_queries": ("keypairs_num_queries", None),
        "projects": ("groups_name", agg_to_array),
    }

    @classmethod
    async def load_count(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        domain_name: Optional[str] = None,
        email: Optional[str] = None,
        is_active: Optional[bool] = None,
        filter: Optional[str] = None,
    ) -> int:
        from .group import association_groups_users, groups
        from .user import users

        j = (
            sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
            .join(association_groups_users, users.c.uuid == association_groups_users.c.user_id)
            .join(groups, association_groups_users.c.group_id == groups.c.id)
        )
        query = sa.select([sa.func.count()]).group_by(keypairs.c.access_key).select_from(j)
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if email is not None:
            query = query.where(keypairs.c.user_id == email)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        async with graph_ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return len(result.all())

    @classmethod
    async def load_slice(
        cls,
        graph_ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        domain_name: Optional[str] = None,
        email: Optional[str] = None,
        is_active: Optional[bool] = None,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> Sequence[KeyPair]:
        from .group import association_groups_users, groups
        from .user import users

        j = (
            sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
            .join(association_groups_users, users.c.uuid == association_groups_users.c.user_id)
            .join(groups, association_groups_users.c.group_id == groups.c.id)
        )
        query = (
            sa.select([
                keypairs,
                users.c.email,
                users.c.full_name,
                agg_to_array(groups.c.name).label("groups_name"),
            ])
            .select_from(j)
            .group_by(keypairs, users.c.email, users.c.full_name)
            .limit(limit)
            .offset(offset)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if email is not None:
            query = query.where(keypairs.c.user_id == email)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        if order is not None:
            qoparser = QueryOrderParser(cls._queryorder_colmap)
            query = qoparser.append_ordering(query, order)
        else:
            query = query.order_by(keypairs.c.created_at.desc())
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(graph_ctx, row)) is not None
            ]

    @classmethod
    async def batch_load_by_email(
        cls,
        graph_ctx: GraphQueryContext,
        user_ids: Sequence[uuid.UUID],
        *,
        domain_name: Optional[str] = None,
        is_active: Optional[bool] = None,
    ) -> Sequence[Sequence[Optional[KeyPair]]]:
        from .group import association_groups_users, groups
        from .user import users

        j = (
            sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
            .join(association_groups_users, users.c.uuid == association_groups_users.c.user_id)
            .join(groups, association_groups_users.c.group_id == groups.c.id)
        )
        query = (
            sa.select([
                keypairs,
                users.c.email,
                users.c.full_name,
                agg_to_array(groups.c.name).label("groups_name"),
            ])
            .select_from(j)
            .where(keypairs.c.user_id.in_(user_ids))
            .group_by(keypairs, users.c.email, users.c.full_name)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        async with graph_ctx.db.begin_readonly() as conn:
            return await batch_multiresult(
                graph_ctx,
                conn,
                query,
                cls,
                user_ids,
                lambda row: row["user_id"],
            )

    @classmethod
    async def batch_load_by_ak(
        cls,
        graph_ctx: GraphQueryContext,
        access_keys: Sequence[AccessKey],
        *,
        domain_name: Optional[str] = None,
    ) -> Sequence[Optional[KeyPair]]:
        from .group import association_groups_users, groups
        from .user import users

        j = (
            sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
            .join(association_groups_users, users.c.uuid == association_groups_users.c.user_id)
            .join(groups, association_groups_users.c.group_id == groups.c.id)
        )
        query = (
            sa.select([
                keypairs,
                users.c.email,
                users.c.full_name,
                agg_to_array(groups.c.name).label("groups_name"),
            ])
            .select_from(j)
            .where(keypairs.c.access_key.in_(access_keys))
            .group_by(keypairs, users.c.email, users.c.full_name)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        async with graph_ctx.db.begin_readonly() as conn:
            return await batch_result(
                graph_ctx,
                conn,
                query,
                cls,
                access_keys,
                lambda row: row["access_key"],
            )


class KeyPairList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(KeyPair, required=True)


class KeyPairInput(graphene.InputObjectType):
    is_active = graphene.Boolean(required=False, default_value=True)
    is_admin = graphene.Boolean(required=False, default_value=False)
    resource_policy = graphene.String(required=True)
    concurrency_limit = graphene.Int(required=False)  # deprecated and ignored
    rate_limit = graphene.Int(required=True)

    # When creating, you MUST set all fields.
    # When modifying, set the field to "None" to skip setting the value.


class KeyPairInputTD(TypedDict):
    is_active: bool
    is_admin: bool
    resource_policy: str
    rate_limit: int


class ModifyKeyPairInput(graphene.InputObjectType):
    is_active = graphene.Boolean(required=False)
    is_admin = graphene.Boolean(required=False)
    resource_policy = graphene.String(required=False)
    concurrency_limit = graphene.Int(required=False)  # deprecated and ignored
    rate_limit = graphene.Int(required=False)


class CreateKeyPair(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        user_id = graphene.String(required=True)
        props = KeyPairInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    keypair = graphene.Field(lambda: KeyPair, required=False)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        user_id: str,
        props: KeyPairInput,
    ) -> CreateKeyPair:
        from .user import users  # noqa

        graph_ctx: GraphQueryContext = info.context
        data = cls.prepare_new_keypair(
            user_id,
            {
                "is_active": props.is_active,
                "is_admin": props.is_admin,
                "resource_policy": props.resource_policy,
                "rate_limit": props.rate_limit,
                # props.concurrency_limit is always ignored
            },
        )
        insert_query = sa.insert(keypairs).values(
            **data,
            user=sa.select([users.c.uuid]).where(users.c.email == user_id).as_scalar(),
        )
        return await simple_db_mutate_returning_item(cls, graph_ctx, insert_query, item_cls=KeyPair)

    @classmethod
    def prepare_new_keypair(cls, user_email: str, props: KeyPairInputTD) -> Dict[str, Any]:
        ak, sk = generate_keypair()
        pubkey, privkey = generate_ssh_keypair()
        data = {
            "user_id": user_email,
            "access_key": ak,
            "secret_key": sk,
            "is_active": props["is_active"],
            "is_admin": props["is_admin"],
            "resource_policy": props["resource_policy"],
            "rate_limit": props["rate_limit"],
            "num_queries": 0,
            "ssh_public_key": pubkey,
            "ssh_private_key": privkey,
        }
        return data


class ModifyKeyPair(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        access_key = graphene.String(required=True)
        props = ModifyKeyPairInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        access_key: AccessKey,
        props: ModifyUserInput,
    ) -> ModifyKeyPair:
        ctx: GraphQueryContext = info.context
        data: Dict[str, Any] = {}
        set_if_set(props, data, "is_active")
        set_if_set(props, data, "is_admin")
        set_if_set(props, data, "resource_policy")
        set_if_set(props, data, "rate_limit")
        # props.concurrency_limit is always ignored
        update_query = sa.update(keypairs).values(data).where(keypairs.c.access_key == access_key)
        return await simple_db_mutate(cls, ctx, update_query)


class DeleteKeyPair(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        access_key = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        access_key: AccessKey,
    ) -> DeleteKeyPair:
        from .user import UserRow

        ctx: GraphQueryContext = info.context
        async with ctx.db.begin_readonly_session() as db_session:
            user_query = (
                sa.select([sa.func.count()])
                .select_from(UserRow)
                .where(UserRow.main_access_key == access_key)
            )
            if (await db_session.scalar(user_query)) > 0:
                return DeleteKeyPair(False, "the keypair is used as main access key by any user")
        delete_query = sa.delete(keypairs).where(keypairs.c.access_key == access_key)
        result = await simple_db_mutate(cls, ctx, delete_query)
        if result.ok:
            await redis_helper.execute(
                ctx.redis_stat,
                lambda r: r.delete(f"keypair.concurrency_used.{access_key}"),
            )
        return result


class Dotfile(TypedDict):
    data: str
    path: str
    perm: str


def generate_keypair() -> Tuple[AccessKey, SecretKey]:
    """
    AWS-like access key and secret key generation.
    """
    ak = "AKIA" + base64.b32encode(secrets.token_bytes(10)).decode("ascii")
    sk = secrets.token_urlsafe(30)
    return AccessKey(ak), SecretKey(sk)


def generate_ssh_keypair() -> Tuple[str, str]:
    """
    Generate RSA keypair for SSH/SFTP connection.
    """
    key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=2048,
    )
    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.TraditionalOpenSSL,
        crypto_serialization.NoEncryption(),
    ).decode("utf-8")
    public_key = (
        key.public_key()
        .public_bytes(
            crypto_serialization.Encoding.OpenSSH,
            crypto_serialization.PublicFormat.OpenSSH,
        )
        .decode("utf-8")
    )
    public_key = f"{public_key.rstrip()}\n"
    private_key = f"{private_key.rstrip()}\n"
    return (public_key, private_key)


async def query_owned_dotfiles(
    conn: SAConnection,
    access_key: AccessKey,
) -> Tuple[List[Dotfile], int]:
    query = (
        sa.select([keypairs.c.dotfiles])
        .select_from(keypairs)
        .where(keypairs.c.access_key == access_key)
    )
    packed_dotfile = (await conn.execute(query)).scalar()
    rows = msgpack.unpackb(packed_dotfile)
    return rows, MAXIMUM_DOTFILE_SIZE - len(packed_dotfile)


async def query_bootstrap_script(
    conn: SAConnection,
    access_key: AccessKey,
) -> Tuple[str, int]:
    query = (
        sa.select([keypairs.c.bootstrap_script])
        .select_from(keypairs)
        .where(keypairs.c.access_key == access_key)
    )
    script = (await conn.execute(query)).scalar()
    return script, MAXIMUM_DOTFILE_SIZE - len(script)


def verify_dotfile_name(dotfile: str) -> bool:
    if dotfile in RESERVED_DOTFILES:
        return False
    return True



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, TypedDict

import graphene
import sqlalchemy as sa
from graphene.types.datetime import DateTime as GQLDateTime
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.orm import relationship

from ai.backend.common import msgpack
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import ResourceSlot

from ..defs import RESERVED_DOTFILES
from .base import (
    Base,
    ResourceSlotColumn,
    VFolderHostPermissionColumn,
    batch_result,
    mapper_registry,
    set_if_set,
    simple_db_mutate,
    simple_db_mutate_returning_item,
)
from .scaling_group import ScalingGroup
from .user import UserRole

if TYPE_CHECKING:
    from .gql import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


__all__: Sequence[str] = (
    "domains",
    "DomainRow",
    "Domain",
    "DomainInput",
    "ModifyDomainInput",
    "CreateDomain",
    "ModifyDomain",
    "DeleteDomain",
    "DomainDotfile",
    "MAXIMUM_DOTFILE_SIZE",
    "query_domain_dotfiles",
    "verify_dotfile_name",
)

MAXIMUM_DOTFILE_SIZE = 64 * 1024  # 61 KiB
_rx_slug = re.compile(r"^[a-zA-Z0-9]([a-zA-Z0-9._-]*[a-zA-Z0-9])?$")

domains = sa.Table(
    "domains",
    mapper_registry.metadata,
    sa.Column("name", sa.String(length=64), primary_key=True),
    sa.Column("description", sa.String(length=512)),
    sa.Column("is_active", sa.Boolean, default=True),
    sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    sa.Column(
        "modified_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        onupdate=sa.func.current_timestamp(),
    ),
    # TODO: separate resource-related fields with new domain resource policy table when needed.
    sa.Column("total_resource_slots", ResourceSlotColumn(), default="{}"),
    sa.Column(
        "allowed_vfolder_hosts",
        VFolderHostPermissionColumn(),
        nullable=False,
        default={},
    ),
    sa.Column("allowed_docker_registries", pgsql.ARRAY(sa.String), nullable=False, default="{}"),
    #: Field for synchronization with external services.
    sa.Column("integration_id", sa.String(length=512)),
    # dotfiles column, \x90 means empty list in msgpack
    sa.Column(
        "dotfiles", sa.LargeBinary(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=b"\x90"
    ),
)


class DomainRow(Base):
    __table__ = domains
    sessions = relationship("SessionRow", back_populates="domain")
    users = relationship("UserRow", back_populates="domain")
    groups = relationship("GroupRow", back_populates="domain")
    scaling_groups = relationship(
        "ScalingGroupRow",
        secondary="sgroups_for_domains",
        back_populates="domains",
    )


class Domain(graphene.ObjectType):
    name = graphene.String()
    description = graphene.String()
    is_active = graphene.Boolean()
    created_at = GQLDateTime()
    modified_at = GQLDateTime()
    total_resource_slots = graphene.JSONString()
    allowed_vfolder_hosts = graphene.JSONString()
    allowed_docker_registries = graphene.List(lambda: graphene.String)
    integration_id = graphene.String()

    # Dynamic fields.
    scaling_groups = graphene.List(lambda: graphene.String)

    async def resolve_scaling_groups(self, info: graphene.ResolveInfo) -> Sequence[str]:
        sgroups = await ScalingGroup.load_by_domain(info.context, self.name)
        return [sg.name for sg in sgroups]

    @classmethod
    def from_row(cls, ctx: GraphQueryContext, row: Row) -> Optional[Domain]:
        if row is None:
            return None
        return cls(
            name=row["name"],
            description=row["description"],
            is_active=row["is_active"],
            created_at=row["created_at"],
            modified_at=row["modified_at"],
            total_resource_slots=row["total_resource_slots"].to_json()
            if row["total_resource_slots"] is not None
            else {},
            allowed_vfolder_hosts=row["allowed_vfolder_hosts"].to_json(),
            allowed_docker_registries=row["allowed_docker_registries"],
            integration_id=row["integration_id"],
        )

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
        *,
        is_active: Optional[bool] = None,
    ) -> Sequence[Domain]:
        async with ctx.db.begin_readonly() as conn:
            query = sa.select([domains]).select_from(domains)
            if is_active is not None:
                query = query.where(domains.c.is_active == is_active)
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(ctx, row)) is not None
            ]

    @classmethod
    async def batch_load_by_name(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
        *,
        is_active: Optional[bool] = None,
    ) -> Sequence[Optional[Domain]]:
        async with ctx.db.begin_readonly() as conn:
            query = sa.select([domains]).select_from(domains).where(domains.c.name.in_(names))
            if is_active is not None:
                query = query.where(domains.c.is_active == is_active)
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                names,
                lambda row: row["name"],
            )


class DomainInput(graphene.InputObjectType):
    description = graphene.String(required=False, default_value="")
    is_active = graphene.Boolean(required=False, default_value=True)
    total_resource_slots = graphene.JSONString(required=False, default_value={})
    allowed_vfolder_hosts = graphene.JSONString(required=False, default_value={})
    allowed_docker_registries = graphene.List(
        lambda: graphene.String, required=False, default_value=[]
    )
    integration_id = graphene.String(required=False, default_value=None)


class ModifyDomainInput(graphene.InputObjectType):
    name = graphene.String(required=False)
    description = graphene.String(required=False)
    is_active = graphene.Boolean(required=False)
    total_resource_slots = graphene.JSONString(required=False)
    allowed_vfolder_hosts = graphene.JSONString(required=False)
    allowed_docker_registries = graphene.List(lambda: graphene.String, required=False)
    integration_id = graphene.String(required=False)


class CreateDomain(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = DomainInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    domain = graphene.Field(lambda: Domain, required=False)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
        props: DomainInput,
    ) -> CreateDomain:
        if _rx_slug.search(name) is None:
            return cls(False, "invalid name format. slug format required.", None)
        ctx: GraphQueryContext = info.context
        data = {
            "name": name,
            "description": props.description,
            "is_active": props.is_active,
            "total_resource_slots": ResourceSlot.from_user_input(props.total_resource_slots, None),
            "allowed_vfolder_hosts": props.allowed_vfolder_hosts,
            "allowed_docker_registries": props.allowed_docker_registries,
            "integration_id": props.integration_id,
        }
        insert_query = sa.insert(domains).values(data)
        return await simple_db_mutate_returning_item(cls, ctx, insert_query, item_cls=Domain)


class ModifyDomain(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = ModifyDomainInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    domain = graphene.Field(lambda: Domain, required=False)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
        props: ModifyDomainInput,
    ) -> ModifyDomain:
        ctx: GraphQueryContext = info.context
        data: Dict[str, Any] = {}
        set_if_set(props, data, "name")  # data['name'] is new domain name
        set_if_set(props, data, "description")
        set_if_set(props, data, "is_active")
        set_if_set(
            props,
            data,
            "total_resource_slots",
            clean_func=lambda v: ResourceSlot.from_user_input(v, None),
        )
        set_if_set(props, data, "allowed_vfolder_hosts")
        set_if_set(props, data, "allowed_docker_registries")
        set_if_set(props, data, "integration_id")
        if "name" in data and _rx_slug.search(data["name"]) is None:
            raise ValueError("invalid name format. slug format required.")
        update_query = sa.update(domains).values(data).where(domains.c.name == name)
        return await simple_db_mutate_returning_item(cls, ctx, update_query, item_cls=Domain)


class DeleteDomain(graphene.Mutation):
    """
    Instead of deleting the domain, just mark it as inactive.
    """

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(cls, root, info: graphene.ResolveInfo, name: str) -> DeleteDomain:
        ctx: GraphQueryContext = info.context
        update_query = sa.update(domains).values(is_active=False).where(domains.c.name == name)
        return await simple_db_mutate(cls, ctx, update_query)


class PurgeDomain(graphene.Mutation):
    """
    Completely delete domain from DB.

    Domain-bound kernels will also be all deleted.
    To purge domain, there should be no users and groups in the target domain.
    """

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(cls, root, info: graphene.ResolveInfo, name: str) -> PurgeDomain:
        from . import groups, users

        ctx: GraphQueryContext = info.context

        async def _pre_func(conn: SAConnection) -> None:
            if await cls.domain_has_active_kernels(conn, name):
                raise RuntimeError("Domain has some active kernels. Terminate them first.")
            query = sa.select([sa.func.count()]).where(users.c.domain_name == name)
            user_count = await conn.scalar(query)
            if user_count > 0:
                raise RuntimeError("There are users bound to the domain. Remove users first.")
            query = sa.select([sa.func.count()]).where(groups.c.domain_name == name)
            group_count = await conn.scalar(query)
            if group_count > 0:
                raise RuntimeError("There are groups bound to the domain. Remove groups first.")

            await cls.delete_kernels(conn, name)

        delete_query = sa.delete(domains).where(domains.c.name == name)
        return await simple_db_mutate(cls, ctx, delete_query, pre_func=_pre_func)

    @classmethod
    async def delete_kernels(
        cls,
        conn: SAConnection,
        domain_name: str,
    ) -> int:
        """
        Delete all kernels run from the target domain.

        :param conn: DB connection
        :param domain_name: domain's name to delete kernels

        :return: number of deleted rows
        """
        from . import kernels

        delete_query = sa.delete(kernels).where(kernels.c.domain_name == domain_name)
        result = await conn.execute(delete_query)
        if result.rowcount > 0:
            log.info('deleted {0} domain"s kernels ({1})', result.rowcount, domain_name)
        return result.rowcount

    @classmethod
    async def domain_has_active_kernels(
        cls,
        conn: SAConnection,
        domain_name: str,
    ) -> bool:
        """
        Check if the domain does not have active kernels.

        :param conn: DB connection
        :param domain_name: domain's name

        :return: True if the domain has some active kernels.
        """
        from . import AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES, kernels

        query = (
            sa.select([sa.func.count()])
            .select_from(kernels)
            .where(
                (kernels.c.domain_name == domain_name)
                & (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES))
            )
        )
        active_kernel_count = await conn.scalar(query)
        return active_kernel_count > 0


class DomainDotfile(TypedDict):
    data: str
    path: str
    perm: str


async def query_domain_dotfiles(
    conn: SAConnection,
    name: str,
) -> tuple[List[DomainDotfile], int]:
    query = sa.select([domains.c.dotfiles]).select_from(domains).where(domains.c.name == name)
    packed_dotfile = await conn.scalar(query)
    if packed_dotfile is None:
        return [], MAXIMUM_DOTFILE_SIZE
    rows = msgpack.unpackb(packed_dotfile)
    return rows, MAXIMUM_DOTFILE_SIZE - len(packed_dotfile)


def verify_dotfile_name(dotfile: str) -> bool:
    if dotfile in RESERVED_DOTFILES:
        return False
    return True



"""change_endpoint_to_nullable

Revision ID: 589c764a18f1
Revises: 3f47af213b05
Create Date: 2024-02-27 20:18:55.524946

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID, convention

# revision identifiers, used by Alembic.
revision = "589c764a18f1"
down_revision = "3f47af213b05"
branch_labels = None
depends_on = None


metadata = sa.MetaData(naming_convention=convention)


BATCH_SIZE = 100


def upgrade():
    op.alter_column("endpoints", "model", nullable=True)
    op.drop_constraint(
        "fk_endpoint_tokens_endpoint_endpoints", "endpoint_tokens", type_="foreignkey"
    )
    op.create_foreign_key(
        op.f("fk_endpoint_tokens_endpoint_endpoints"),
        "endpoint_tokens",
        "endpoints",
        ["endpoint"],
        ["id"],
        ondelete="SET NULL",
    )
    op.drop_constraint("fk_endpoints_model_vfolders", "endpoints", type_="foreignkey")
    op.create_foreign_key(
        op.f("fk_endpoints_model_vfolders"),
        "endpoints",
        "vfolders",
        ["model"],
        ["id"],
        ondelete="SET NULL",
    )


def downgrade():
    conn = op.get_bind()

    endpoint_tokens = sa.Table(
        "endpoint_tokens",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column(
            "endpoint", GUID, sa.ForeignKey("endpoints.id", ondelete="SET NULL"), nullable=True
        ),
        extend_existing=True,
    )
    endpoints = sa.Table(
        "endpoints",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("model", GUID, sa.ForeignKey("vfolders.id", ondelete="SET NULL"), nullable=True),
        extend_existing=True,
    )

    def _delete(table, null_field):
        while True:
            subq = sa.select([table.c.id]).where(null_field.is_(sa.null())).limit(BATCH_SIZE)
            delete_stmt = sa.delete(table).where(table.c.id.in_(subq))
            result = conn.execute(delete_stmt)
            if result.rowcount == 0:
                break

    _delete(endpoint_tokens, endpoint_tokens.c.endpoint)
    _delete(endpoints, endpoints.c.model)

    op.drop_constraint(op.f("fk_endpoints_model_vfolders"), "endpoints", type_="foreignkey")
    op.create_foreign_key(
        "fk_endpoints_model_vfolders",
        "endpoints",
        "vfolders",
        ["model"],
        ["id"],
        ondelete="RESTRICT",
    )
    op.alter_column("endpoints", "model", nullable=False)
    op.drop_constraint(
        op.f("fk_endpoint_tokens_endpoint_endpoints"), "endpoint_tokens", type_="foreignkey"
    )
    op.create_foreign_key(
        "fk_endpoint_tokens_endpoint_endpoints",
        "endpoint_tokens",
        "endpoints",
        ["endpoint"],
        ["id"],
    )



"""add_session_table

Revision ID: b6b884fbae1f
Revises: cace152eefac
Create Date: 2022-12-05 16:12:53.275671

"""

from collections import defaultdict
from typing import Any, Mapping, Sequence
from uuid import UUID, uuid4

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.orm import registry

from ai.backend.manager.defs import DEFAULT_ROLE
from ai.backend.manager.models import KernelStatus, SessionStatus
from ai.backend.manager.models.base import GUID, KernelIDColumn, convention
from ai.backend.manager.models.session import SessionDependencyRow

# revision identifiers, used by Alembic.
revision = "b6b884fbae1f"
down_revision = "cace152eefac"
branch_labels = None
depends_on = None


kernel_status = tuple(s.name for s in KernelStatus)

session_status = tuple(s.name for s in SessionStatus)

metadata = sa.MetaData(naming_convention=convention)
mapper_registry = registry(metadata=metadata)
Base = mapper_registry.generate_base()

PAGE_SIZE = 100


def default_hostname(context) -> str:
    params = context.get_current_parameters()
    return f"{params['cluster_role']}{params['cluster_idx']}"


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    connection = op.get_bind()
    kernels = sa.Table(
        "kernels",
        metadata,
        KernelIDColumn(),
        sa.Column("session_id", GUID, nullable=False),
        sa.Column("session_creation_id", sa.String(length=32), unique=False),
        sa.Column("session_name", sa.String(length=64), unique=False),
        sa.Column(
            "session_type",
            pgsql.ENUM("INTERACTIVE", "BATCH"),
            nullable=False,
            default="INTERACTIVE",
            server_default="INTERACTIVE",
        ),
        sa.Column("cluster_role", sa.String(length=16), nullable=False, default="main"),
        sa.Column(
            "cluster_mode",
            sa.String(length=16),
            nullable=False,
            default="SINGLE_NODE",
            server_default="SINGLE_NODE",
        ),
        sa.Column("cluster_size", sa.Integer, nullable=False, default=1),
        sa.Column("cluster_idx", sa.Integer, nullable=False, default=0),
        sa.Column("local_rank", sa.Integer, nullable=False, default=0),
        sa.Column(
            "cluster_hostname", sa.String(length=64), nullable=False, default=default_hostname
        ),
        # Resource ownership
        sa.Column("agent", sa.String(length=64), sa.ForeignKey("agents.id"), nullable=True),
        sa.Column("agent_addr", sa.String(length=128), nullable=True),
        sa.Column("scaling_group", sa.ForeignKey("scaling_groups.name"), nullable=True),
        sa.Column(
            "domain_name", sa.String(length=64), sa.ForeignKey("domains.name"), nullable=False
        ),
        sa.Column("group_id", GUID, sa.ForeignKey("groups.id"), nullable=False),
        sa.Column("user_uuid", GUID, sa.ForeignKey("users.uuid"), nullable=False),
        sa.Column("access_key", sa.String(length=20), sa.ForeignKey("keypairs.access_key")),
        sa.Column("image", sa.String(length=512)),
        sa.Column("architecture", sa.String(length=32), default="x86_64"),
        sa.Column("registry", sa.String(length=512)),
        sa.Column("tag", sa.String(length=64), nullable=True),
        # Resource occupation
        sa.Column("container_id", sa.String(length=64)),
        sa.Column("occupied_slots", pgsql.JSONB(), nullable=False),
        sa.Column("occupied_shares", pgsql.JSONB(), nullable=False, default={}),  # legacy
        sa.Column("environ", sa.ARRAY(sa.String), nullable=True),
        sa.Column("mounts", sa.ARRAY(sa.String), nullable=True),  # list of list; legacy since 22.03
        sa.Column("mount_map", pgsql.JSONB(), nullable=True, default={}),  # legacy since 22.03
        sa.Column("vfolder_mounts", pgsql.JSONB(), nullable=True),
        sa.Column("attached_devices", pgsql.JSONB(), nullable=True, default={}),
        sa.Column("resource_opts", pgsql.JSONB(), nullable=True, default={}),
        sa.Column("bootstrap_script", sa.String(length=16 * 1024), nullable=True),
        # Port mappings
        # If kernel_host is NULL, it is assumed to be same to the agent host or IP.
        sa.Column("kernel_host", sa.String(length=128), nullable=True),
        sa.Column("repl_in_port", sa.Integer(), nullable=False),
        sa.Column("repl_out_port", sa.Integer(), nullable=False),
        sa.Column("stdin_port", sa.Integer(), nullable=False),  # legacy for stream_pty
        sa.Column("stdout_port", sa.Integer(), nullable=False),  # legacy for stream_pty
        sa.Column("service_ports", pgsql.JSONB(), nullable=True),
        sa.Column("preopen_ports", sa.ARRAY(sa.Integer), nullable=True),
        sa.Column("use_host_network", sa.Boolean(), default=False, nullable=False),
        # Lifecycle
        sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
        sa.Column("terminated_at", sa.DateTime(timezone=True), nullable=True, default=sa.null()),
        sa.Column("starts_at", sa.DateTime(timezone=True), nullable=True, default=sa.null()),
        sa.Column(
            "status",
            sa.Enum(*kernel_status),
            default="PENDING",
            server_default="PENDING",
            nullable=False,
        ),
        sa.Column("status_changed", sa.DateTime(timezone=True), nullable=True),
        sa.Column("status_history", pgsql.JSONB(), nullable=True, default=sa.null()),
        sa.Column("status_info", sa.Unicode(), nullable=True, default=sa.null()),
        sa.Column("status_data", pgsql.JSONB(), nullable=True, default=sa.null()),
        sa.Column("callback_url", sa.types.UnicodeText, nullable=True, default=sa.null()),
        sa.Column("startup_command", sa.Text, nullable=True),
        sa.Column(
            "result",
            sa.Enum("UNDEFINED", "SUCCESS", "FAILURE"),
            default="UNDEFINED",
            server_default="UNDEFINED",
            nullable=False,
        ),
        sa.Column("internal_data", pgsql.JSONB(), nullable=True),
        sa.Column("container_log", sa.LargeBinary(), nullable=True),
        # Resource metrics measured upon termination
        sa.Column("num_queries", sa.BigInteger(), default=0),
        sa.Column("last_stat", pgsql.JSONB(), nullable=True, default=sa.null()),
    )

    class SessionRow(Base):  # type: ignore[valid-type, misc]
        __tablename__ = "sessions"
        id = sa.Column(
            "id",
            GUID(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
            primary_key=True,
        )
        creation_id = sa.Column("creation_id", sa.String(length=32), unique=False)
        name = sa.Column("name", sa.String(length=64), unique=False)
        session_type = sa.Column(
            "session_type",
            pgsql.ENUM("INTERACTIVE", "BATCH", name="sessiontypes", create_type=False),
            server_default="INTERACTIVE",
            nullable=False,
        )
        cluster_mode = sa.Column(
            "cluster_mode", sa.String(length=16), server_default="SINGLE_NODE", nullable=False
        )
        cluster_size = sa.Column("cluster_size", sa.Integer(), nullable=False)

        # Resource ownership
        scaling_group_name = sa.Column("scaling_group_name", sa.String(length=64), nullable=True)
        target_sgroup_names = sa.Column(
            "target_sgroup_names",
            sa.ARRAY(sa.String(length=64)),
            server_default="{}",
            nullable=True,
        )
        domain_name = sa.Column("domain_name", sa.String(length=64), nullable=False)
        group_id = sa.Column("group_id", GUID(), nullable=False)
        user_uuid = sa.Column("user_uuid", GUID(), nullable=False)
        access_key = sa.Column("access_key", sa.String(length=20), nullable=True)
        # kp_resource_policy = sa.Column(
        #     'kp_resource_policy', sa.String(length=256), nullable=False)

        # # if image_id is null, should find a image field from related kernel row.
        # image_id = sa.Column("image_id", GUID(), nullable=True)
        # # `image` column is identical to kernels `image` column.
        # image = sa.Column("image", sa.String(length=512))
        tag = sa.Column("tag", sa.String(length=64), nullable=True)

        # Resource occupation
        occupying_slots = sa.Column("occupying_slots", pgsql.JSONB(), nullable=False)
        requested_slots = sa.Column("requested_slots", pgsql.JSONB(), nullable=False)
        vfolder_mounts = sa.Column("vfolder_mounts", pgsql.JSONB(), nullable=True)
        resource_opts = sa.Column("resource_opts", pgsql.JSONB(), nullable=True, default={})
        environ = sa.Column("environ", pgsql.JSONB(), nullable=True)
        bootstrap_script = sa.Column("bootstrap_script", sa.String(length=16 * 1024), nullable=True)
        use_host_network = sa.Column(
            "use_host_network", sa.Boolean(), default=False, nullable=False
        )

        # Lifecycle
        timeout = sa.Column("timeout", sa.BigInteger(), nullable=True)
        created_at = sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        )
        terminated_at = sa.Column("terminated_at", sa.DateTime(timezone=True), nullable=True)
        starts_at = sa.Column("starts_at", sa.DateTime(timezone=True), nullable=True)
        status = sa.Column(
            "status",
            pgsql.ENUM(
                *session_status,
                name="sessionstatus",
                create_type=True,
            ),
            server_default="PENDING",
            nullable=False,
        )
        # status_changed = sa.Column("status_changed", sa.DateTime(timezone=True), nullable=True)
        status_history = sa.Column(
            "status_history", pgsql.JSONB(), nullable=True, default=sa.null()
        )
        status_info = sa.Column("status_info", sa.Unicode(), nullable=True)
        status_data = sa.Column("status_data", pgsql.JSONB(astext_type=sa.Text()), nullable=True)
        callback_url = sa.Column("callback_url", sa.types.UnicodeText(), nullable=True)

        startup_command = sa.Column("startup_command", sa.Text(), nullable=True)
        result = sa.Column(
            "result",
            pgsql.ENUM("UNDEFINED", "SUCCESS", "FAILURE", name="sessionresults", create_type=True),
            server_default="UNDEFINED",
            nullable=False,
        )

        # Resource metrics measured upon termination
        num_queries = sa.Column("num_queries", sa.BigInteger(), nullable=True)
        last_stat = sa.Column("last_stat", pgsql.JSONB(astext_type=sa.Text()), nullable=True)

    SessionRow.__table__.create(connection)
    op.create_foreign_key(
        op.f("fk_sessions_access_key_keypairs"),
        "sessions",
        "keypairs",
        ["access_key"],
        ["access_key"],
    )
    op.create_foreign_key(
        op.f("fk_sessions_domain_name_domains"), "sessions", "domains", ["domain_name"], ["name"]
    )
    op.create_foreign_key(
        op.f("fk_sessions_group_id_groups"), "sessions", "groups", ["group_id"], ["id"]
    )
    # op.create_foreign_key(
    #     op.f("fk_sessions_image_images"), "sessions", "images", ["image_id"], ["id"]
    # )
    op.create_foreign_key(
        op.f("fk_sessions_scaling_group_scaling_groups"),
        "sessions",
        "scaling_groups",
        ["scaling_group_name"],
        ["name"],
    )
    op.create_foreign_key(
        op.f("fk_sessions_user_uuid_users"), "sessions", "users", ["user_uuid"], ["uuid"]
    )
    op.create_index(op.f("ix_sessions_created_at"), "sessions", ["created_at"], unique=False)
    op.create_index(op.f("ix_sessions_name"), "sessions", ["name"], unique=False)
    op.create_index(op.f("ix_sessions_result"), "sessions", ["result"], unique=False)
    op.create_index(
        op.f("ix_sessions_scaling_group_name"), "sessions", ["scaling_group_name"], unique=False
    )
    op.create_index(op.f("ix_sessions_session_type"), "sessions", ["session_type"], unique=False)
    # op.create_index(op.f("ix_sessions_status"), "sessions", ["status"], unique=False)
    # op.create_index(
    #     op.f("ix_sessions_status_changed"), "sessions", ["status_changed"], unique=False
    # )
    op.create_index(op.f("ix_sessions_terminated_at"), "sessions", ["terminated_at"], unique=False)

    op.drop_constraint(
        "fk_session_dependencies_depends_on_kernels", "session_dependencies", type_="foreignkey"
    )
    op.drop_constraint(
        "fk_session_dependencies_session_id_kernels", "session_dependencies", type_="foreignkey"
    )

    def migrate_kernel_to_session(kernel_query: sa.sql.Select, is_single_kernel: bool) -> None:
        session_map: dict[UUID, dict[str, Any]] = {}
        main_kernel_rows: dict[UUID, dict[str, Any]] = {}
        single_kernel_ids: dict[UUID, UUID] = {}
        kernel_rows: Sequence[Mapping[str, Any]] = connection.execute(kernel_query).fetchall()

        def _map_session(session_id: UUID, kernel_row: Mapping[str, Any]) -> dict[str, Any]:
            return {
                "id": session_id,
                "creation_id": kernel_row["session_creation_id"],
                "name": kernel_row["session_name"],
                "session_type": kernel_row["session_type"],
                "cluster_mode": kernel_row["cluster_mode"],
                "cluster_size": kernel_row["cluster_size"],
                "scaling_group_name": kernel_row["scaling_group"],
                "domain_name": kernel_row["domain_name"],
                "group_id": kernel_row["group_id"],
                "user_uuid": kernel_row["user_uuid"],
                "access_key": kernel_row["access_key"],
                # "image_id": img_map[row["image"]],
                # "image": row["image"],
                "tag": kernel_row["tag"],
                "occupying_slots": kernel_row["occupied_slots"],
                "requested_slots": kernel_row["occupied_slots"],
                "vfolder_mounts": kernel_row["vfolder_mounts"],
                "resource_opts": kernel_row["resource_opts"],
                "bootstrap_script": kernel_row["bootstrap_script"],
                "use_host_network": kernel_row["use_host_network"],
                "created_at": kernel_row["created_at"],
                "terminated_at": kernel_row["terminated_at"],
                "starts_at": kernel_row["starts_at"],
                "status": kernel_row["status"],
                # "status_changed": row["status_changed"],
                "status_history": kernel_row["status_history"],
                "status_info": kernel_row["status_info"],
                "status_data": kernel_row["status_data"],
                "callback_url": kernel_row["callback_url"],
                "startup_command": kernel_row["startup_command"],
                "result": kernel_row["result"],
                "num_queries": kernel_row["num_queries"],
                "last_stat": kernel_row["last_stat"],
            }

        if is_single_kernel:
            for row in kernel_rows:
                sess_id = uuid4()
                session_map[sess_id] = _map_session(sess_id, row)
                single_kernel_ids[row["id"]] = sess_id
        else:
            for row in kernel_rows:
                # Since main kernel of multi-kernel session has the same id with session_id
                # insert new kernels to kernels have distinguished id with sessions.
                if row["cluster_role"] == "main" or row["cluster_role"] == DEFAULT_ROLE:
                    new_kern_id = uuid4()
                    main_kernel_rows[row["id"]] = {
                        **row,
                        "id": new_kern_id,
                    }

                sess_id = row["session_id"]
                if sess_id not in session_map:
                    session_map[sess_id] = _map_session(sess_id, row)
                else:
                    sess = session_map[sess_id]
                    try:
                        if sess["created_at"] is not None and row["created_at"] is not None:
                            if sess["created_at"] > row["created_at"]:
                                sess["created_at"] = row["created_at"]
                                # st_change = row["st_change"]
                    except (TypeError, ValueError):
                        pass

                    try:
                        if sess["starts_at"] is not None and row["starts_at"] is not None:
                            if sess["starts_at"] > row["starts_at"]:
                                sess["starts_at"] = row["starts_at"]
                                # st_change = row["st_change"]
                    except (TypeError, ValueError):
                        pass

                    # if str(sess["status"]) not in ("ERROR", "CANCELLED"):
                    #     if KernelStatus.index(str(sess["status"])) > KernelStatus.index(str(row["status"])):
                    #         sess["status"] = row["status"]
                    sess["status"] = row["status"]

                    try:
                        if sess["terminated_at"] is not None and row["terminated_at"] is not None:
                            if sess["terminated_at"] < row["terminated_at"]:
                                sess["terminated_at"] = row["terminated_at"]
                                # st_change = row["status_changed"]
                        elif sess["terminated_at"] is None:
                            pass
                        elif row["terminated_at"] is None:
                            sess["terminated_at"] = None
                        # sess["status_changed"] = st_change
                    except (TypeError, ValueError):
                        pass

                    if sess["status_info"] != row["status_info"]:
                        if row["cluster_role"] == "main" or row["cluster_role"] == DEFAULT_ROLE:
                            sess["status_info"] = row["status_info"]

        if session_map:
            creates = tuple(session_map.values())
            connection.execute(SessionRow.__table__.insert(), creates)

        if is_single_kernel and single_kernel_ids:
            # Delete and insert session_dependencies
            # Since session_id and depends_on columns have primary constraint, we should not `update` them.
            kern_ids = list(single_kernel_ids.keys())
            insert_values = []
            sess_dep_query = sa.select(SessionDependencyRow).where(
                SessionDependencyRow.session_id.in_(kern_ids)
                | SessionDependencyRow.depends_on.in_(kern_ids)
            )
            for row in connection.execute(sess_dep_query).fetchall():
                val = {}
                if row["session_id"] in single_kernel_ids:
                    val["session_id"] = single_kernel_ids[row["session_id"]]
                else:
                    val["session_id"] = row["session_id"]

                if row["depends_on"] in single_kernel_ids:
                    val["depends_on"] = single_kernel_ids[row["depends_on"]]
                else:
                    val["depends_on"] = row["depends_on"]
                insert_values.append(val)
            dep_delete_query = sa.delete(SessionDependencyRow).where(
                SessionDependencyRow.session_id.in_(kern_ids)
                | SessionDependencyRow.depends_on.in_(kern_ids)
            )
            connection.execute(dep_delete_query)
            if insert_values:
                connection.execute(sa.insert(SessionDependencyRow), insert_values)

            # Update single-kernel session's kernel
            sess_query = (
                sa.update(kernels)
                .where(kernels.c.id == sa.bindparam("kernel_id"))
                .values(session_id=sa.bindparam("b_session_id"))
            )
            sess_update_params = [
                {
                    "kernel_id": kern_id,
                    "b_session_id": sess_id,
                }
                for kern_id, sess_id in single_kernel_ids.items()
            ]
            if sess_update_params:
                connection.execute(sess_query, sess_update_params)

        # Insert and delete multi-kernel session's main kernel
        if main_kernel_rows:
            connection.execute(sa.insert(kernels), list(main_kernel_rows.values()))
            delete_query = sa.delete(kernels).where(kernels.c.id.in_(list(main_kernel_rows.keys())))
            connection.execute(delete_query)

    # Drop the unique constraint before migration
    op.drop_index("ix_kernels_unique_sess_token", table_name="kernels")

    # Multi-kernel session migration
    session_cnt = connection.execute(
        sa.select([sa.func.count(sa.distinct(kernels.c.session_id))]).where(
            kernels.c.cluster_size > 1
        )
    ).scalar()
    for offset in range(0, session_cnt, PAGE_SIZE):
        session_id_query = (
            sa.select([sa.distinct(kernels.c.session_id)])
            .where(kernels.c.cluster_size > 1)
            .order_by(kernels.c.session_id)
            .offset(offset)
            .limit(PAGE_SIZE)
        )
        session_ids = connection.execute(session_id_query).scalars().all()
        query = sa.select([kernels]).where(kernels.c.session_id.in_(session_ids))
        migrate_kernel_to_session(query, is_single_kernel=False)

    # Single-kernel session migration
    kernel_cnt = connection.execute(
        sa.select([sa.func.count()]).select_from(kernels).where(kernels.c.cluster_size == 1)
    ).scalar()
    for offset in range(0, kernel_cnt, PAGE_SIZE):
        query = (
            sa.select([kernels])
            .where(kernels.c.cluster_size == 1)
            .order_by(kernels.c.id)
            .offset(offset)
            .limit(PAGE_SIZE)
        )
        migrate_kernel_to_session(query, is_single_kernel=True)

    # Restore the dropped constraint.
    op.create_index(
        "ix_kernels_unique_sess_token",
        "kernels",
        ["access_key", "session_name"],
        unique=True,
        postgresql_where=sa.text(
            "status NOT IN ('TERMINATED', 'CANCELLED') and cluster_role = 'main'"
        ),
    )

    op.create_foreign_key(
        op.f("fk_session_dependencies_session_id_sessions"),
        "session_dependencies",
        "sessions",
        ["session_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        op.f("fk_session_dependencies_depends_on_sessions"),
        "session_dependencies",
        "sessions",
        ["depends_on"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )

    op.create_foreign_key(
        op.f("fk_kernels_session_id_sessions"), "kernels", "sessions", ["session_id"], ["id"]
    )
    op.add_column("kernels", sa.Column("requested_slots", pgsql.JSONB(), nullable=True))

    ## Association tables
    # groups_users
    op.drop_constraint(
        "uq_association_user_id_group_id", "association_groups_users", type_="unique"
    )

    # scaling_groups_domains
    op.drop_constraint("uq_sgroup_domain", "sgroups_for_domains", type_="unique")

    # scaling_groups_groups
    op.drop_constraint("uq_sgroup_ugroup", "sgroups_for_groups", type_="unique")

    # scaling_groups_keypairs
    op.drop_constraint("uq_sgroup_akey", "sgroups_for_keypairs", type_="unique")

    # ### end Alembic commands ###


def downgrade() -> None:
    connection = op.get_bind()

    # Kernel table
    op.drop_constraint(op.f("fk_kernels_session_id_sessions"), "kernels", type_="foreignkey")
    kernels = sa.Table(
        "kernels",
        metadata,
        KernelIDColumn(),
        sa.Column("session_id", GUID, nullable=False),
        sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    )
    op.drop_constraint(
        op.f("fk_session_dependencies_depends_on_sessions"),
        "session_dependencies",
        type_="foreignkey",
    )
    op.drop_constraint(
        op.f("fk_session_dependencies_session_id_sessions"),
        "session_dependencies",
        type_="foreignkey",
    )

    class SessionRow(Base):  # type: ignore[valid-type, misc]
        __tablename__ = "sessions"
        id = sa.Column(
            "id",
            GUID(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
            primary_key=True,
        )

    session_cnt = connection.execute(sa.select([sa.func.count()]).select_from(SessionRow)).scalar()
    for page in range(0, session_cnt, PAGE_SIZE):
        query = sa.select(SessionRow.id).order_by(SessionRow.id).offset(page).limit(PAGE_SIZE)
        session_ids = [row["id"] for row in connection.execute(query).fetchall()]

        query = sa.select([kernels]).where(kernels.c.session_id.in_(session_ids))
        kernel_rows = connection.execute(query).fetchall()

        num_kernel_sess: dict[UUID, int] = defaultdict(int)
        sess_kern_map = {}
        for kern in kernel_rows:
            # restore kernel's session_id
            sess_id = kern["session_id"]
            num_kernel_sess[sess_id] += 1
            if sess_id not in sess_kern_map:
                sess_kern_map[sess_id] = kern["id"]
        kern_update_query = (
            sa.update(kernels)
            .where(kernels.c.id == sa.bindparam("kernel_id"))
            .values({kernels.c.session_id: sa.bindparam("session_id")})
        )
        sess_update_params = []
        single_kern_sess = {}
        for sess_id, num_kern in num_kernel_sess.items():
            if num_kern == 1:
                # it is single kernel session.
                # session_id must be same with kernel_id.
                kernel_id = sess_kern_map[sess_id]
                single_kern_sess[sess_id] = kernel_id
                sess_update_params.append({
                    "kernel_id": kernel_id,
                    "session_id": kernel_id,
                })
        if sess_update_params:
            connection.execute(kern_update_query, sess_update_params)

        # Session dependency table
        sess_ids = list(single_kern_sess.keys())
        sess_dep_query = sa.select(SessionDependencyRow).where(
            SessionDependencyRow.session_id.in_(sess_ids)
            | SessionDependencyRow.depends_on.in_(sess_ids)
        )
        insert_values = []
        for row in connection.execute(sess_dep_query).fetchall():
            val = {}
            if row["session_id"] in single_kern_sess:
                val["session_id"] = single_kern_sess[row["session_id"]]
            else:
                val["session_id"] = row["session_id"]

            if row["depends_on"] in single_kern_sess:
                val["depends_on"] = single_kern_sess[row["depends_on"]]
            else:
                val["depends_on"] = row["depends_on"]
            insert_values.append(val)
        dep_delete_query = sa.delete(SessionDependencyRow).where(
            SessionDependencyRow.session_id.in_(sess_ids)
            | SessionDependencyRow.depends_on.in_(sess_ids)
        )
        connection.execute(dep_delete_query)
        if insert_values:
            connection.execute(sa.insert(SessionDependencyRow), insert_values)

    # op.drop_column("kernels", "image_id")
    op.drop_column("kernels", "requested_slots")
    op.create_foreign_key(
        "fk_session_dependencies_session_id_kernels",
        "session_dependencies",
        "kernels",
        ["session_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "fk_session_dependencies_depends_on_kernels",
        "session_dependencies",
        "kernels",
        ["depends_on"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )

    # Session table
    op.drop_index(op.f("ix_sessions_terminated_at"), table_name="sessions")
    # op.drop_index(op.f("ix_sessions_status_changed"), table_name="sessions")
    # op.drop_index(op.f("ix_sessions_status"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_session_type"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_scaling_group_name"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_result"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_name"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_created_at"), table_name="sessions")
    op.drop_table("sessions")

    # Drop ENUM types
    sess_statues = pgsql.ENUM(
        *session_status,
        name="sessionstatus",
    )
    sess_results = pgsql.ENUM("UNDEFINED", "SUCCESS", "FAILURE", name="sessionresults")

    sess_statues.drop(op.get_bind())
    sess_results.drop(op.get_bind())

    ## Association tables
    # groups users
    op.create_unique_constraint(
        "uq_association_user_id_group_id", "association_groups_users", ["user_id", "group_id"]
    )

    # scaling groups domains
    op.create_unique_constraint(
        "uq_sgroup_domain", "sgroups_for_domains", ["scaling_group", "domain"]
    )

    # scaling_groups_groups
    op.create_unique_constraint(
        "uq_sgroup_ugroup", "sgroups_for_groups", ["scaling_group", "group"]
    )

    # scaling_groups_keypairs
    op.create_unique_constraint(
        "uq_sgroup_akey", "sgroups_for_keypairs", ["scaling_group", "access_key"]
    )

    ### end Alembic commands ###



from __future__ import annotations

import uuid
from datetime import timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Set,
    cast,
    overload,
)

import attr
import graphene
import sqlalchemy as sa
import trafaret as t
from graphene.types.datetime import DateTime as GQLDateTime
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.orm import load_only, relationship
from sqlalchemy.sql.expression import true

from ai.backend.common import validators as tx
from ai.backend.common.types import (
    AgentSelectionStrategy,
    JSONSerializableMixin,
    ResourceSlot,
    SessionTypes,
)

from .agent import AgentStatus
from .base import (
    Base,
    IDColumn,
    StructuredJSONObjectColumn,
    batch_multiresult,
    batch_result,
    mapper_registry,
    set_if_set,
    simple_db_mutate,
    simple_db_mutate_returning_item,
)
from .group import resolve_group_name_or_id, resolve_groups
from .user import UserRole

if TYPE_CHECKING:
    from .gql import GraphQueryContext

__all__: Sequence[str] = (
    # table defs
    "scaling_groups",
    "ScalingGroupOpts",
    "ScalingGroupRow",
    "sgroups_for_domains",
    "sgroups_for_groups",
    "sgroups_for_keypairs",
    # functions
    "query_allowed_sgroups",
    "ScalingGroup",
    "CreateScalingGroup",
    "ModifyScalingGroup",
    "DeleteScalingGroup",
    "AssociateScalingGroupWithDomain",
    "AssociateScalingGroupWithUserGroup",
    "AssociateScalingGroupWithKeyPair",
    "DisassociateScalingGroupWithDomain",
    "DisassociateScalingGroupWithUserGroup",
    "DisassociateScalingGroupWithKeyPair",
)


@attr.define(slots=True)
class ScalingGroupOpts(JSONSerializableMixin):
    allowed_session_types: list[SessionTypes] = attr.Factory(
        lambda: [
            SessionTypes.INTERACTIVE,
            SessionTypes.BATCH,
            SessionTypes.INFERENCE,
        ],
    )
    pending_timeout: timedelta = timedelta(seconds=0)
    config: Mapping[str, Any] = attr.Factory(dict)
    agent_selection_strategy: AgentSelectionStrategy = AgentSelectionStrategy.DISPERSED
    roundrobin: bool = False

    def to_json(self) -> dict[str, Any]:
        return {
            "allowed_session_types": [item.value for item in self.allowed_session_types],
            "pending_timeout": self.pending_timeout.total_seconds(),
            "config": self.config,
            "agent_selection_strategy": self.agent_selection_strategy,
            "roundrobin": self.roundrobin,
        }

    @classmethod
    def from_json(cls, obj: Mapping[str, Any]) -> ScalingGroupOpts:
        return cls(**cls.as_trafaret().check(obj))

    @classmethod
    def as_trafaret(cls) -> t.Trafaret:
        return t.Dict({
            t.Key("allowed_session_types", default=["interactive", "batch"]): t.List(
                tx.Enum(SessionTypes), min_length=1
            ),
            t.Key("pending_timeout", default=0): tx.TimeDuration(allow_negative=False),
            # Each scheduler impl refers an additional "config" key.
            t.Key("config", default={}): t.Mapping(t.String, t.Any),
            t.Key("agent_selection_strategy", default=AgentSelectionStrategy.DISPERSED): tx.Enum(
                AgentSelectionStrategy
            ),
            t.Key("roundrobin", default=False): t.Bool(),
        }).allow_extra("*")


scaling_groups = sa.Table(
    "scaling_groups",
    mapper_registry.metadata,
    sa.Column("name", sa.String(length=64), primary_key=True),
    sa.Column("description", sa.String(length=512)),
    sa.Column("is_active", sa.Boolean, index=True, default=True),
    sa.Column(
        "is_public", sa.Boolean, index=True, default=True, server_default=true(), nullable=False
    ),
    sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    sa.Column("wsproxy_addr", sa.String(length=1024), nullable=True),
    sa.Column("wsproxy_api_token", sa.String(length=128), nullable=True),
    sa.Column("driver", sa.String(length=64), nullable=False),
    sa.Column("driver_opts", pgsql.JSONB(), nullable=False, default={}),
    sa.Column("scheduler", sa.String(length=64), nullable=False),
    sa.Column("use_host_network", sa.Boolean, nullable=False, default=False),
    sa.Column(
        "scheduler_opts",
        StructuredJSONObjectColumn(ScalingGroupOpts),
        nullable=False,
        default={},
    ),
)


# When scheduling, we take the union of allowed scaling groups for
# each domain, group, and keypair.


sgroups_for_domains = sa.Table(
    "sgroups_for_domains",
    mapper_registry.metadata,
    IDColumn(),
    sa.Column(
        "scaling_group",
        sa.ForeignKey("scaling_groups.name", onupdate="CASCADE", ondelete="CASCADE"),
        index=True,
        nullable=False,
    ),
    sa.Column(
        "domain",
        sa.ForeignKey("domains.name", onupdate="CASCADE", ondelete="CASCADE"),
        index=True,
        nullable=False,
    ),
    sa.UniqueConstraint("scaling_group", "domain", name="uq_sgroup_domain"),
)


sgroups_for_groups = sa.Table(
    "sgroups_for_groups",
    mapper_registry.metadata,
    IDColumn(),
    sa.Column(
        "scaling_group",
        sa.ForeignKey("scaling_groups.name", onupdate="CASCADE", ondelete="CASCADE"),
        index=True,
        nullable=False,
    ),
    sa.Column(
        "group",
        sa.ForeignKey("groups.id", onupdate="CASCADE", ondelete="CASCADE"),
        index=True,
        nullable=False,
    ),
    sa.UniqueConstraint("scaling_group", "group", name="uq_sgroup_ugroup"),
)


sgroups_for_keypairs = sa.Table(
    "sgroups_for_keypairs",
    mapper_registry.metadata,
    IDColumn(),
    sa.Column(
        "scaling_group",
        sa.ForeignKey("scaling_groups.name", onupdate="CASCADE", ondelete="CASCADE"),
        index=True,
        nullable=False,
    ),
    sa.Column(
        "access_key",
        sa.ForeignKey("keypairs.access_key", onupdate="CASCADE", ondelete="CASCADE"),
        index=True,
        nullable=False,
    ),
    sa.UniqueConstraint("scaling_group", "access_key", name="uq_sgroup_akey"),
)


class ScalingGroupRow(Base):
    __table__ = scaling_groups
    sessions = relationship("SessionRow", back_populates="scaling_group")
    agents = relationship("AgentRow", back_populates="scaling_group_row")
    domains = relationship(
        "DomainRow",
        secondary=sgroups_for_domains,
        back_populates="scaling_groups",
    )
    groups = relationship(
        "GroupRow",
        secondary=sgroups_for_groups,
        back_populates="scaling_groups",
    )
    keypairs = relationship(
        "KeyPairRow",
        secondary=sgroups_for_keypairs,
        back_populates="scaling_groups",
    )


@overload
async def query_allowed_sgroups(
    db_conn: SAConnection,
    domain_name: str,
    group: uuid.UUID,
    access_key: str,
) -> Sequence[Row]: ...


@overload
async def query_allowed_sgroups(
    db_conn: SAConnection,
    domain_name: str,
    group: Iterable[uuid.UUID],
    access_key: str,
) -> Sequence[Row]: ...


@overload
async def query_allowed_sgroups(
    db_conn: SAConnection,
    domain_name: str,
    group: str,
    access_key: str,
) -> Sequence[Row]: ...


@overload
async def query_allowed_sgroups(
    db_conn: SAConnection,
    domain_name: str,
    group: Iterable[str],
    access_key: str,
) -> Sequence[Row]: ...


async def query_allowed_sgroups(
    db_conn: SAConnection,
    domain_name: str,
    group: uuid.UUID | Iterable[uuid.UUID] | str | Iterable[str],
    access_key: str,
) -> Sequence[Row]:
    query = sa.select([sgroups_for_domains]).where(sgroups_for_domains.c.domain == domain_name)
    result = await db_conn.execute(query)
    from_domain = {row["scaling_group"] for row in result}

    group_ids: Iterable[uuid.UUID] = []
    match group:
        case uuid.UUID() | str():
            if group_id := await resolve_group_name_or_id(db_conn, domain_name, group):
                group_ids = [group_id]
            else:
                group_ids = []
        case list() | tuple() | set():
            group_ids = await resolve_groups(db_conn, domain_name, cast(Iterable, group))
    from_group: Set[str]
    if not group_ids:
        from_group = set()  # empty
    else:
        group_cond = sgroups_for_groups.c.group.in_(group_ids)
        query = sa.select([sgroups_for_groups]).where(group_cond)
        result = await db_conn.execute(query)
        from_group = {row["scaling_group"] for row in result}

    query = sa.select([sgroups_for_keypairs]).where(sgroups_for_keypairs.c.access_key == access_key)
    result = await db_conn.execute(query)
    from_keypair = {row["scaling_group"] for row in result}

    sgroups = from_domain | from_group | from_keypair
    query = (
        sa.select([scaling_groups])
        .where(
            (scaling_groups.c.name.in_(sgroups)) & (scaling_groups.c.is_active),
        )
        .order_by(scaling_groups.c.name)
    )
    result = await db_conn.execute(query)
    return [row for row in result]


class ScalingGroup(graphene.ObjectType):
    name = graphene.String()
    description = graphene.String()
    is_active = graphene.Boolean()
    is_public = graphene.Boolean()
    created_at = GQLDateTime()
    wsproxy_addr = graphene.String()
    wsproxy_api_token = graphene.String()
    driver = graphene.String()
    driver_opts = graphene.JSONString()
    scheduler = graphene.String()
    scheduler_opts = graphene.JSONString()
    use_host_network = graphene.Boolean()

    # Dynamic fields.
    agent_count_by_status = graphene.Field(
        graphene.Int,
        description="Added in 24.03.9.",
        status=graphene.String(
            default_value=AgentStatus.ALIVE.name,
            description=f"Possible states of an agent. Should be one of {[s.name for s in AgentStatus]}. Default is 'ALIVE'.",
        ),
    )

    agent_total_resource_slots_by_status = graphene.Field(
        graphene.JSONString,
        description="Added in 24.03.9.",
        status=graphene.String(
            default_value=AgentStatus.ALIVE.name,
            description=f"Possible states of an agent. Should be one of {[s.name for s in AgentStatus]}. Default is 'ALIVE'.",
        ),
    )

    async def resolve_agent_count_by_status(
        self, info: graphene.ResolveInfo, status: str = AgentStatus.ALIVE.name
    ) -> int:
        from .agent import Agent

        return await Agent.load_count(
            info.context,
            raw_status=status,
            scaling_group=self.name,
        )

    async def resolve_agent_total_resource_slots_by_status(
        self, info: graphene.ResolveInfo, status: str = AgentStatus.ALIVE.name
    ) -> Mapping[str, Any]:
        from .agent import AgentRow, AgentStatus

        graph_ctx = info.context
        async with graph_ctx.db.begin_readonly_session() as db_session:
            query_stmt = (
                sa.select(AgentRow)
                .where(
                    (AgentRow.scaling_group == self.name) & (AgentRow.status == AgentStatus[status])
                )
                .options(load_only(AgentRow.occupied_slots, AgentRow.available_slots))
            )
            result = (await db_session.scalars(query_stmt)).all()
            agent_rows = cast(list[AgentRow], result)

            total_occupied_slots = ResourceSlot()
            total_available_slots = ResourceSlot()

            for agent_row in agent_rows:
                total_occupied_slots += agent_row.occupied_slots
                total_available_slots += agent_row.available_slots

            return {
                "occupied_slots": total_occupied_slots.to_json(),
                "available_slots": total_available_slots.to_json(),
            }

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row | None,
    ) -> ScalingGroup | None:
        if row is None:
            return None
        return cls(
            name=row["name"],
            description=row["description"],
            is_active=row["is_active"],
            is_public=row["is_public"],
            created_at=row["created_at"],
            wsproxy_addr=row["wsproxy_addr"],
            wsproxy_api_token=row["wsproxy_api_token"],
            driver=row["driver"],
            driver_opts=row["driver_opts"],
            scheduler=row["scheduler"],
            scheduler_opts=row["scheduler_opts"].to_json(),
            use_host_network=row["use_host_network"],
        )

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
        *,
        is_active: Optional[bool] = None,
    ) -> Sequence[ScalingGroup]:
        query = sa.select([scaling_groups]).select_from(scaling_groups)
        if is_active is not None:
            query = query.where(scaling_groups.c.is_active == is_active)
        async with ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(ctx, row)) is not None
            ]

    @classmethod
    async def load_by_domain(
        cls,
        ctx: GraphQueryContext,
        domain: str,
        *,
        is_active: Optional[bool] = None,
    ) -> Sequence[ScalingGroup]:
        j = sa.join(
            scaling_groups,
            sgroups_for_domains,
            scaling_groups.c.name == sgroups_for_domains.c.scaling_group,
        )
        query = (
            sa.select([scaling_groups]).select_from(j).where(sgroups_for_domains.c.domain == domain)
        )
        if is_active is not None:
            query = query.where(scaling_groups.c.is_active == is_active)
        async with ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(ctx, row)) is not None
            ]

    @classmethod
    async def load_by_group(
        cls,
        ctx: GraphQueryContext,
        group: uuid.UUID,
        *,
        is_active: Optional[bool] = None,
    ) -> Sequence[ScalingGroup]:
        j = sa.join(
            scaling_groups,
            sgroups_for_groups,
            scaling_groups.c.name == sgroups_for_groups.c.scaling_group,
        )
        query = (
            sa.select([scaling_groups]).select_from(j).where(sgroups_for_groups.c.group == group)
        )
        if is_active is not None:
            query = query.where(scaling_groups.c.is_active == is_active)
        async with ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(ctx, row)) is not None
            ]

    @classmethod
    async def load_by_keypair(
        cls,
        ctx: GraphQueryContext,
        access_key: str,
        *,
        is_active: Optional[bool] = None,
    ) -> Sequence[ScalingGroup]:
        j = sa.join(
            scaling_groups,
            sgroups_for_keypairs,
            scaling_groups.c.name == sgroups_for_keypairs.c.scaling_group,
        )
        query = (
            sa.select([scaling_groups])
            .select_from(j)
            .where(sgroups_for_keypairs.c.access_key == access_key)
        )
        if is_active is not None:
            query = query.where(scaling_groups.c.is_active == is_active)
        async with ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(ctx, row)) is not None
            ]

    @classmethod
    async def batch_load_by_group(
        cls,
        ctx: GraphQueryContext,
        group_ids: Sequence[uuid.UUID],
    ) -> Sequence[Sequence[ScalingGroup | None]]:
        j = sa.join(
            scaling_groups,
            sgroups_for_groups,
            scaling_groups.c.name == sgroups_for_groups.c.scaling_group,
        )
        query = (
            sa.select([scaling_groups, sgroups_for_groups.c.group])
            .select_from(j)
            .where(sgroups_for_groups.c.group.in_(group_ids))
        )
        async with ctx.db.begin_readonly() as conn:
            return await batch_multiresult(
                ctx,
                conn,
                query,
                cls,
                group_ids,
                lambda row: row["group"],
            )

    @classmethod
    async def batch_load_by_name(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
    ) -> Sequence[ScalingGroup | None]:
        query = (
            sa.select([scaling_groups])
            .select_from(scaling_groups)
            .where(scaling_groups.c.name.in_(names))
        )
        async with ctx.db.begin_readonly() as conn:
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                names,
                lambda row: row["name"],
            )


class CreateScalingGroupInput(graphene.InputObjectType):
    description = graphene.String(required=False, default_value="")
    is_active = graphene.Boolean(required=False, default_value=True)
    is_public = graphene.Boolean(required=False, default_value=True)
    wsproxy_addr = graphene.String(required=False, default_value=None)
    wsproxy_api_token = graphene.String(required=False, default_value=None)
    driver = graphene.String(required=True)
    driver_opts = graphene.JSONString(required=False, default_value={})
    scheduler = graphene.String(required=True)
    scheduler_opts = graphene.JSONString(required=False, default_value={})
    use_host_network = graphene.Boolean(required=False, default_value=False)


class ModifyScalingGroupInput(graphene.InputObjectType):
    description = graphene.String(required=False)
    is_active = graphene.Boolean(required=False)
    is_public = graphene.Boolean(required=False)
    wsproxy_addr = graphene.String(required=False)
    wsproxy_api_token = graphene.String(required=False)
    driver = graphene.String(required=False)
    driver_opts = graphene.JSONString(required=False)
    scheduler = graphene.String(required=False)
    scheduler_opts = graphene.JSONString(required=False)
    use_host_network = graphene.Boolean(required=False)


class CreateScalingGroup(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = CreateScalingGroupInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    scaling_group = graphene.Field(lambda: ScalingGroup, required=False)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
        props: CreateScalingGroupInput,
    ) -> CreateScalingGroup:
        data = {
            "name": name,
            "description": props.description,
            "is_active": bool(props.is_active),
            "is_public": bool(props.is_public),
            "wsproxy_addr": props.wsproxy_addr,
            "wsproxy_api_token": props.wsproxy_api_token,
            "driver": props.driver,
            "driver_opts": props.driver_opts,
            "scheduler": props.scheduler,
            "scheduler_opts": ScalingGroupOpts.from_json(props.scheduler_opts),
            "use_host_network": bool(props.use_host_network),
        }
        insert_query = sa.insert(scaling_groups).values(data)
        return await simple_db_mutate_returning_item(
            cls,
            info.context,
            insert_query,
            item_cls=ScalingGroup,
        )


class ModifyScalingGroup(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = ModifyScalingGroupInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
        props: ModifyScalingGroupInput,
    ) -> ModifyScalingGroup:
        data: Dict[str, Any] = {}
        set_if_set(props, data, "description")
        set_if_set(props, data, "is_active")
        set_if_set(props, data, "is_public")
        set_if_set(props, data, "wsproxy_addr")
        set_if_set(props, data, "wsproxy_api_token")
        set_if_set(props, data, "driver")
        set_if_set(props, data, "driver_opts")
        set_if_set(props, data, "scheduler")
        set_if_set(
            props, data, "scheduler_opts", clean_func=lambda v: ScalingGroupOpts.from_json(v)
        )
        set_if_set(props, data, "use_host_network")
        update_query = sa.update(scaling_groups).values(data).where(scaling_groups.c.name == name)
        return await simple_db_mutate(cls, info.context, update_query)


class DeleteScalingGroup(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
    ) -> DeleteScalingGroup:
        delete_query = sa.delete(scaling_groups).where(scaling_groups.c.name == name)
        return await simple_db_mutate(cls, info.context, delete_query)


class AssociateScalingGroupWithDomain(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_group = graphene.String(required=True)
        domain = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_group: str,
        domain: str,
    ) -> AssociateScalingGroupWithDomain:
        insert_query = sa.insert(sgroups_for_domains).values({
            "scaling_group": scaling_group,
            "domain": domain,
        })
        return await simple_db_mutate(cls, info.context, insert_query)


class AssociateScalingGroupsWithDomain(graphene.Mutation):
    """Added in 24.03.9."""

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_groups = graphene.List(graphene.String, required=True)
        domain = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_groups: Sequence[str],
        domain: str,
    ) -> AssociateScalingGroupsWithDomain:
        insert_query = sa.insert(sgroups_for_domains).values([
            {"scaling_group": scaling_group, "domain": domain} for scaling_group in scaling_groups
        ])
        return await simple_db_mutate(cls, info.context, insert_query)


class DisassociateScalingGroupWithDomain(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_group = graphene.String(required=True)
        domain = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_group: str,
        domain: str,
    ) -> DisassociateScalingGroupWithDomain:
        delete_query = sa.delete(sgroups_for_domains).where(
            (sgroups_for_domains.c.scaling_group == scaling_group)
            & (sgroups_for_domains.c.domain == domain),
        )
        return await simple_db_mutate(cls, info.context, delete_query)


class DisassociateScalingGroupsWithDomain(graphene.Mutation):
    """Added in 24.03.9."""

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_groups = graphene.List(graphene.String, required=True)
        domain = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_groups: Sequence[str],
        domain: str,
    ) -> DisassociateScalingGroupsWithDomain:
        delete_query = sa.delete(sgroups_for_domains).where(
            (sgroups_for_domains.c.scaling_group.in_(scaling_groups))
            & (sgroups_for_domains.c.domain == domain),
        )
        return await simple_db_mutate(cls, info.context, delete_query)


class DisassociateAllScalingGroupsWithDomain(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        domain = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        domain: str,
    ) -> DisassociateAllScalingGroupsWithDomain:
        delete_query = sa.delete(sgroups_for_domains).where(sgroups_for_domains.c.domain == domain)
        return await simple_db_mutate(cls, info.context, delete_query)


class AssociateScalingGroupWithUserGroup(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_group = graphene.String(required=True)
        user_group = graphene.UUID(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_group: str,
        user_group: uuid.UUID,
    ) -> AssociateScalingGroupWithUserGroup:
        insert_query = sa.insert(sgroups_for_groups).values({
            "scaling_group": scaling_group,
            "group": user_group,
        })
        return await simple_db_mutate(cls, info.context, insert_query)


class AssociateScalingGroupsWithUserGroup(graphene.Mutation):
    """Added in 24.03.9."""

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_groups = graphene.List(graphene.String, required=True)
        user_group = graphene.UUID(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_groups: Sequence[str],
        user_group: uuid.UUID,
    ) -> AssociateScalingGroupsWithUserGroup:
        insert_query = sa.insert(sgroups_for_groups).values([
            {"scaling_group": scaling_group, "group": user_group}
            for scaling_group in scaling_groups
        ])
        return await simple_db_mutate(cls, info.context, insert_query)


class DisassociateScalingGroupWithUserGroup(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_group = graphene.String(required=True)
        user_group = graphene.UUID(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_group: str,
        user_group: uuid.UUID,
    ) -> DisassociateScalingGroupWithUserGroup:
        delete_query = sa.delete(sgroups_for_groups).where(
            (sgroups_for_groups.c.scaling_group == scaling_group)
            & (sgroups_for_groups.c.group == user_group),
        )
        return await simple_db_mutate(cls, info.context, delete_query)


class DisassociateScalingGroupsWithUserGroup(graphene.Mutation):
    """Added in 24.03.9."""

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_groups = graphene.List(graphene.String, required=True)
        user_group = graphene.UUID(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_groups: Sequence[str],
        user_group: uuid.UUID,
    ) -> DisassociateScalingGroupsWithUserGroup:
        delete_query = sa.delete(sgroups_for_groups).where(
            (sgroups_for_groups.c.scaling_group.in_(scaling_groups))
            & (sgroups_for_groups.c.group == user_group),
        )
        return await simple_db_mutate(cls, info.context, delete_query)


class DisassociateAllScalingGroupsWithGroup(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        user_group = graphene.UUID(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        user_group: uuid.UUID,
    ) -> DisassociateAllScalingGroupsWithGroup:
        delete_query = sa.delete(sgroups_for_groups).where(sgroups_for_groups.c.group == user_group)
        return await simple_db_mutate(cls, info.context, delete_query)


class AssociateScalingGroupWithKeyPair(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_group = graphene.String(required=True)
        access_key = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_group: str,
        access_key: str,
    ) -> AssociateScalingGroupWithKeyPair:
        insert_query = sa.insert(sgroups_for_keypairs).values({
            "scaling_group": scaling_group,
            "access_key": access_key,
        })
        return await simple_db_mutate(cls, info.context, insert_query)


class AssociateScalingGroupsWithKeyPair(graphene.Mutation):
    """Added in 24.03.9."""

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_groups = graphene.List(graphene.String, required=True)
        access_key = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_groups: Sequence[str],
        access_key: str,
    ) -> AssociateScalingGroupsWithKeyPair:
        insert_query = sa.insert(sgroups_for_keypairs).values([
            {"scaling_group": scaling_group, "access_key": access_key}
            for scaling_group in scaling_groups
        ])
        return await simple_db_mutate(cls, info.context, insert_query)


class DisassociateScalingGroupWithKeyPair(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_group = graphene.String(required=True)
        access_key = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_group: str,
        access_key: str,
    ) -> DisassociateScalingGroupWithKeyPair:
        delete_query = sa.delete(sgroups_for_keypairs).where(
            (sgroups_for_keypairs.c.scaling_group == scaling_group)
            & (sgroups_for_keypairs.c.access_key == access_key),
        )
        return await simple_db_mutate(cls, info.context, delete_query)


class DisassociateScalingGroupsWithKeyPair(graphene.Mutation):
    """Added in 24.03.9."""

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        scaling_groups = graphene.List(graphene.String, required=True)
        access_key = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        scaling_groups: Sequence[str],
        access_key: str,
    ) -> DisassociateScalingGroupsWithKeyPair:
        delete_query = sa.delete(sgroups_for_keypairs).where(
            (sgroups_for_keypairs.c.scaling_group.in_(scaling_groups))
            & (sgroups_for_keypairs.c.access_key == access_key),
        )
        return await simple_db_mutate(cls, info.context, delete_query)



import enum
from typing import Any, Callable, Mapping, Optional, Type, TypeAlias, TypeVar, Union

import sqlalchemy as sa
from lark import Lark, LarkError, Transformer, Tree
from lark.lexer import Token

from . import ArrayFieldItem, EnumFieldItem, FieldSpecItem, JSONFieldItem, get_col_from_table

__all__ = (
    "FieldSpecType",
    "FilterableSQLQuery",
    "QueryFilterParser",
)

_grammar = r"""
    ?start: expr
    value: string
           | number
           | array
           | ATOM -> atom
    ATOM       : "null" | "true" | "false"
    COMBINE_OP : "&" | "|"
    UNARY_OP   : "!"
    BINARY_OP  : "==" | "!="
               | ">" | ">="
               | "<" | "<="
               | "contains" | "in"
               | "isnot" | "is"
               | "like" | "ilike"
    expr: UNARY_OP expr         -> unary_expr
        | CNAME BINARY_OP value -> binary_expr
        | expr COMBINE_OP expr  -> combine_expr
        | "(" expr ")"          -> paren_expr
    array  : "[" [value ("," value)*] "]"
    string : ESCAPED_STRING
    number : SIGNED_NUMBER
    %import common.CNAME
    %import common.ESCAPED_STRING
    %import common.SIGNED_NUMBER
    %import common.WS
    %ignore WS
"""
_parser = Lark(
    _grammar,
    parser="lalr",
    maybe_placeholders=False,
)

FilterableSQLQuery = Union[sa.sql.Select, sa.sql.Update, sa.sql.Delete]
FieldSpecType: TypeAlias = Mapping[str, FieldSpecItem] | None
WhereClauseType: TypeAlias = (
    sa.sql.expression.BinaryExpression | sa.sql.expression.BooleanClauseList
)
T_Enum = TypeVar("T_Enum", bound=enum.Enum)


def enum_field_getter(enum_cls: Type[T_Enum]) -> Callable[[str], T_Enum]:
    def get_enum(value: str) -> T_Enum:
        for enum_name in (value, value.upper()):
            try:
                return enum_cls[enum_name]
            except KeyError:
                continue
        else:
            enum_names = ", ".join([e.name for e in enum_cls])
            raise ValueError(f"expected one of `{enum_names}` or lower names; got `{value}`")

    return get_enum


class QueryFilterTransformer(Transformer):
    def __init__(self, sa_table: sa.Table, fieldspec: Optional[FieldSpecType] = None) -> None:
        super().__init__()
        self._sa_table = sa_table
        self._fieldspec = fieldspec

    def string(self, token: list[Token]) -> str:
        s = token[0]
        # SQL-side escaping is handled by SQLAlchemy
        return s[1:-1].replace('\\"', '"')

    def number(self, token: list[Token]) -> int | float:
        n = token[0]
        if "." in n:
            return float(n)
        return int(n)

    array = list

    def atom(self, token: list[Token]) -> Type[sa.sql.elements.SingletonConstant]:
        a = token[0]
        if a.value == "null":
            return sa.null()
        elif a.value == "true":
            return sa.true()
        elif a.value == "false":
            return sa.false()
        raise ValueError("Unknown/unsupported atomic token", a.value)

    def _transform_val_leaf(self, col_name: str, op: str, value: Any) -> Any:
        if self._fieldspec:
            try:
                func = self._fieldspec[col_name][1]
            except KeyError:
                raise ValueError("Unknown/unsupported field name", col_name)
            return func(value) if func is not None else value
        else:
            return value

    def _transform_val(self, col_name: str, op: str, value: Any) -> Any:
        if isinstance(value, Tree):
            val = self._transform_val(col_name, op, value.children[0])
        elif isinstance(value, list):
            val = [self._transform_val(col_name, op, v) for v in value]
        else:
            val = self._transform_val_leaf(col_name, op, value)
        return val

    def binary_expr(self, *args) -> sa.sql.elements.BinaryExpression:
        children: list[Token] = args[0]
        col_name = children[0].value
        op = children[1].value
        val = self._transform_val(col_name, op, children[2])

        def build_expr(op: str, col, val):
            match op:
                case "==":
                    expr = col == val
                case "!=":
                    expr = col != val
                case ">":
                    expr = col > val
                case ">=":
                    expr = col >= val
                case "<":
                    expr = col < val
                case "<=":
                    expr = col <= val
                case "contains":
                    expr = col.contains(val)
                case "in":
                    expr = col.in_(val)
                case "isnot":
                    expr = col.isnot(val)
                case "is":
                    expr = col.is_(val)
                case "like":
                    expr = col.like(val)
                case "ilike":
                    expr = col.ilike(val)
                case _:
                    expr = args
            return expr

        try:
            if self._fieldspec is not None:
                match self._fieldspec[children[0].value][0]:
                    case ArrayFieldItem(col_name):
                        # For array columns, let's apply the expression on every item,
                        # and select the row if anyone makes the result true.
                        col = get_col_from_table(self._sa_table, col_name)
                        unnested_col = sa.func.unnest(col).alias("item")
                        subq = (
                            sa.select([sa.column("item")])
                            .select_from(unnested_col)
                            .where(build_expr(op, sa.column("item"), val))
                        )
                        expr = sa.exists(subq)
                    case JSONFieldItem(col_name, obj_key):
                        # For json columns, we additionally indicate the object key
                        # to retrieve the value used in the expression.
                        col = get_col_from_table(self._sa_table, col_name).op("->>")(obj_key)
                        expr = build_expr(op, col, val)
                    case EnumFieldItem(col_name, enum_cls):
                        col = get_col_from_table(self._sa_table, col_name)
                        # allow both key and value of enum to be specified on variable `val`
                        # fetch original enum pointer from given `val`
                        try:
                            enum_val = enum_cls(val)
                        except ValueError:
                            try:
                                enum_val = enum_cls[val]
                            except KeyError:
                                raise ValueError(f"Invalid enum value: {val}")
                        expr = build_expr(op, col, enum_val)
                    case str(col_name):
                        col = get_col_from_table(self._sa_table, col_name)
                        expr = build_expr(op, col, val)
            else:
                col = get_col_from_table(self._sa_table, col_name)
                expr = build_expr(op, col, val)
        except KeyError:
            raise ValueError("Unknown/unsupported field name", col_name)
        return expr

    def unary_expr(self, *args):
        children = args[0]
        op = children[0].value
        expr = children[1]
        if op in ("not", "!"):
            return sa.not_(expr)
        return args

    def combine_expr(self, *args):
        children = args[0]
        op = children[1].value
        expr1 = children[0]
        expr2 = children[2]
        if op == "&":
            return sa.and_(expr1, expr2)
        elif op == "|":
            return sa.or_(expr1, expr2)
        return args

    def paren_expr(self, *args):
        children = args[0]
        return children[0]


class QueryFilterParser:
    def __init__(self, fieldspec: Optional[FieldSpecType] = None) -> None:
        self._fieldspec = fieldspec
        self._parser = _parser

    def parse_filter(
        self,
        table,
        filter_expr: str,
    ) -> WhereClauseType:
        try:
            ast = self._parser.parse(filter_expr)
            where_clause = QueryFilterTransformer(table, self._fieldspec).transform(ast)
        except LarkError as e:
            raise ValueError(f"Query filter parsing error: {e}")
        return where_clause

    def append_filter(
        self,
        sa_query: FilterableSQLQuery,
        filter_expr: str,
    ) -> FilterableSQLQuery:
        """
        Parse the given filter expression and build the where clause based on the first target table from
        the given SQLAlchemy query object.
        """
        if isinstance(sa_query, sa.sql.Select):
            table = sa_query.froms[0]
        elif isinstance(sa_query, sa.sql.Delete):
            table = sa_query.table
        elif isinstance(sa_query, sa.sql.Update):
            table = sa_query.table
        else:
            raise ValueError("Unsupported SQLAlchemy query object type")
        where_clause = self.parse_filter(table, filter_expr)
        final_query = sa_query.where(where_clause)
        assert final_query is not None
        return final_query



from __future__ import annotations

import asyncio
import enum
import functools
import json
import logging
import socket
from typing import TYPE_CHECKING, Any, Final, FrozenSet, Iterable, Tuple

import aiohttp_cors
import attrs
import graphene
import sqlalchemy as sa
import trafaret as t
from aiohttp import web
from aiotools import aclosing

from ai.backend.common import redis_helper
from ai.backend.common import validators as tx
from ai.backend.common.events import DoPrepareEvent, DoScaleEvent, DoScheduleEvent
from ai.backend.common.logging import BraceStyleAdapter

from .. import __version__
from ..defs import DEFAULT_ROLE
from ..models import AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES, agents, kernels
from . import ManagerStatus, SchedulerEvent
from .auth import superadmin_required
from .exceptions import (
    GenericBadRequest,
    InstanceNotFound,
    InvalidAPIParameters,
    ServerFrozen,
    ServiceUnavailable,
)
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params, set_handler_attr

if TYPE_CHECKING:
    from ai.backend.manager.models.gql import GraphQueryContext

    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class SchedulerOps(enum.Enum):
    INCLUDE_AGENTS = "include-agents"
    EXCLUDE_AGENTS = "exclude-agents"


def server_status_required(allowed_status: FrozenSet[ManagerStatus]):
    def decorator(handler):
        @functools.wraps(handler)
        async def wrapped(request, *args, **kwargs) -> web.StreamResponse:
            root_ctx: RootContext = request.app["_root.context"]
            status = await root_ctx.shared_config.get_manager_status()
            if status not in allowed_status:
                if status == ManagerStatus.FROZEN:
                    raise ServerFrozen
                msg = f"Server is not in the required status: {allowed_status}"
                raise ServiceUnavailable(msg)
            return await handler(request, *args, **kwargs)

        set_handler_attr(wrapped, "server_status_required", True)
        set_handler_attr(wrapped, "required_server_statuses", allowed_status)

        return wrapped

    return decorator


READ_ALLOWED: Final = frozenset({ManagerStatus.RUNNING, ManagerStatus.FROZEN})
ALL_ALLOWED: Final = frozenset({ManagerStatus.RUNNING})


class GQLMutationUnfrozenRequiredMiddleware:
    def resolve(self, next, root, info: graphene.ResolveInfo, **args) -> Any:
        graph_ctx: GraphQueryContext = info.context
        if (
            info.operation.operation == "mutation"
            and graph_ctx.manager_status == ManagerStatus.FROZEN
        ):
            raise ServerFrozen
        return next(root, info, **args)


async def detect_status_update(root_ctx: RootContext) -> None:
    try:
        async with aclosing(root_ctx.shared_config.watch_manager_status()) as agen:
            async for ev in agen:
                if ev.event == "put":
                    root_ctx.shared_config.get_manager_status.cache_clear()
                    updated_status = await root_ctx.shared_config.get_manager_status()
                    log.debug(
                        "Process-{0} detected manager status update: {1}",
                        root_ctx.pidx,
                        updated_status,
                    )
    except asyncio.CancelledError:
        pass


async def fetch_manager_status(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    log.info("MANAGER.FETCH_MANAGER_STATUS ()")
    try:
        status = await root_ctx.shared_config.get_manager_status()
        # etcd_info = await root_ctx.shared_config.get_manager_nodes_info()
        configs = root_ctx.local_config["manager"]

        async with root_ctx.db.begin() as conn:
            query = (
                sa.select([sa.func.count()])
                .select_from(kernels)
                .where(
                    (kernels.c.cluster_role == DEFAULT_ROLE)
                    & (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES)),
                )
            )
            active_sessions_num = await conn.scalar(query)

            _id = configs["id"] if configs.get("id") else socket.gethostname()
            nodes = [
                {
                    "id": _id,
                    "num_proc": configs["num-proc"],
                    "service_addr": str(configs["service-addr"]),
                    "heartbeat_timeout": configs["heartbeat-timeout"],
                    "ssl_enabled": configs["ssl-enabled"],
                    "active_sessions": active_sessions_num,
                    "status": status.value,
                    "version": __version__,
                    "api_version": request["api_version"],
                },
            ]
            return web.json_response({
                "nodes": nodes,
                "status": status.value,  # legacy?
                "active_sessions": active_sessions_num,  # legacy?
            })
    except Exception:
        log.exception("GET_MANAGER_STATUS: exception")
        raise


@superadmin_required
@check_api_params(
    t.Dict({
        t.Key("status"): tx.Enum(ManagerStatus),
        t.Key("force_kill", default=False): t.ToBool,
    })
)
async def update_manager_status(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    log.info(
        "MANAGER.UPDATE_MANAGER_STATUS (status:{}, force_kill:{})",
        params["status"],
        params["force_kill"],
    )
    try:
        status = params["status"]
        force_kill = params["force_kill"]
    except json.JSONDecodeError:
        raise InvalidAPIParameters(extra_msg="No request body!")
    except (AssertionError, ValueError) as e:
        raise InvalidAPIParameters(extra_msg=str(e.args[0]))

    if force_kill:
        await root_ctx.registry.kill_all_sessions()
    await root_ctx.shared_config.update_manager_status(status)

    return web.Response(status=204)


async def get_announcement(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    data = await root_ctx.shared_config.etcd.get("manager/announcement")
    if data is None:
        ret = {"enabled": False, "message": ""}
    else:
        ret = {"enabled": True, "message": data}
    return web.json_response(ret)


@superadmin_required
@check_api_params(
    t.Dict({
        t.Key("enabled", default="false"): t.ToBool,
        t.Key("message", default=None): t.Null | t.String,
    })
)
async def update_announcement(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    if params["enabled"]:
        if not params["message"]:
            raise InvalidAPIParameters(extra_msg="Empty message not allowed to enable announcement")
        await root_ctx.shared_config.etcd.put("manager/announcement", params["message"])
    else:
        await root_ctx.shared_config.etcd.delete("manager/announcement")
    return web.Response(status=204)


iv_scheduler_ops_args = {
    SchedulerOps.INCLUDE_AGENTS: t.List(t.String),
    SchedulerOps.EXCLUDE_AGENTS: t.List(t.String),
}


@superadmin_required
@check_api_params(
    t.Dict({
        t.Key("op"): tx.Enum(SchedulerOps),
        t.Key("args"): t.Any,
    })
)
async def perform_scheduler_ops(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    try:
        args = iv_scheduler_ops_args[params["op"]].check(params["args"])
    except t.DataError as e:
        raise InvalidAPIParameters(
            f"Input validation failed for args with {params['op']}",
            extra_data=e.as_dict(),
        )
    if params["op"] in (SchedulerOps.INCLUDE_AGENTS, SchedulerOps.EXCLUDE_AGENTS):
        schedulable = params["op"] == SchedulerOps.INCLUDE_AGENTS
        async with root_ctx.db.begin() as conn:
            query = agents.update().values(schedulable=schedulable).where(agents.c.id.in_(args))
            result = await conn.execute(query)
            if result.rowcount < len(args):
                raise InstanceNotFound()
        if schedulable:
            # trigger scheduler
            await root_ctx.event_producer.produce_event(DoScheduleEvent())
    else:
        raise GenericBadRequest("Unknown scheduler operation")
    return web.Response(status=204)


@superadmin_required
@check_api_params(
    t.Dict({
        t.Key("event"): tx.Enum(SchedulerEvent),
    })
)
async def scheduler_trigger(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    match params["event"]:
        case SchedulerEvent.SCHEDULE:
            await root_ctx.event_producer.produce_event(DoScheduleEvent())
        case SchedulerEvent.PREPARE:
            await root_ctx.event_producer.produce_event(DoPrepareEvent())
        case SchedulerEvent.SCALE_SERVICES:
            await root_ctx.event_producer.produce_event(DoScaleEvent())
    return web.Response(status=204)


@superadmin_required
async def scheduler_healthcheck(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    manager_id = root_ctx.local_config["manager"]["id"]

    scheduler_status = {}
    for event in SchedulerEvent:
        scheduler_status[event.value] = await redis_helper.execute(
            root_ctx.redis_live,
            lambda r: r.hgetall(f"manager.{manager_id}.{event.value}"),
            encoding="utf-8",
        )

    return web.json_response(scheduler_status)


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    status_watch_task: asyncio.Task


async def init(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["manager.context"]
    app_ctx.status_watch_task = asyncio.create_task(detect_status_update(root_ctx))


async def shutdown(app: web.Application) -> None:
    app_ctx: PrivateContext = app["manager.context"]
    if app_ctx.status_watch_task is not None:
        app_ctx.status_watch_task.cancel()
        await app_ctx.status_watch_task


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["api_versions"] = (2, 3, 4)
    app["manager.context"] = PrivateContext()
    app["prefix"] = "manager"
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    status_resource = cors.add(app.router.add_resource("/status"))
    cors.add(status_resource.add_route("GET", fetch_manager_status))
    cors.add(status_resource.add_route("PUT", update_manager_status))
    announcement_resource = cors.add(app.router.add_resource("/announcement"))
    cors.add(announcement_resource.add_route("GET", get_announcement))
    cors.add(announcement_resource.add_route("POST", update_announcement))
    cors.add(app.router.add_route("POST", "/scheduler/operation", perform_scheduler_ops))
    cors.add(app.router.add_route("POST", "/scheduler/trigger", scheduler_trigger))
    cors.add(app.router.add_route("GET", "/scheduler/status", scheduler_healthcheck))
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    return app, []



from __future__ import annotations

import json
import logging
from contextlib import asynccontextmanager as actxmgr
from typing import AsyncIterator, Optional

import aiohttp
import sqlalchemy as sa
import yarl

from ai.backend.common.docker import get_docker_connector
from ai.backend.common.logging import BraceStyleAdapter

from ..models.image import ImageRow
from .base import (
    BaseContainerRegistry,
    concurrency_sema,
    progress_reporter,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class LocalRegistry(BaseContainerRegistry):
    @actxmgr
    async def prepare_client_session(self) -> AsyncIterator[tuple[yarl.URL, aiohttp.ClientSession]]:
        connector = get_docker_connector()
        async with aiohttp.ClientSession(connector=connector.connector) as sess:
            yield connector.docker_host, sess

    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        async with sess.get(self.registry_url / "images" / "json") as response:
            items = await response.json()
            if (reporter := progress_reporter.get()) is not None:
                reporter.total_progress = len(items)
            for item in items:
                labels = item["Labels"]
                if not labels:
                    continue
                if "ai.backend.kernelspec" in labels and item["RepoTags"] is not None:
                    for image_ref_str in item["RepoTags"]:
                        if image_ref_str == "<none>:<none>":
                            # cache images
                            continue
                        yield image_ref_str  # this includes the tag part

    async def _scan_image(
        self,
        sess: aiohttp.ClientSession,
        image: str,
    ) -> None:
        repo, _, tag = image.rpartition(":")
        await self._scan_tag_local(sess, {}, repo, tag)

    async def _scan_tag_local(
        self,
        sess: aiohttp.ClientSession,
        rqst_args: dict[str, str],
        image: str,
        digest: str,
        tag: Optional[str] = None,
    ) -> None:
        async def _read_image_info(
            _tag: str,
        ) -> tuple[dict[str, dict], str | None]:
            async with sess.get(
                self.registry_url / "images" / f"{image}:{digest}" / "json"
            ) as response:
                data = await response.json()
            architecture = data["Architecture"]
            summary = {
                "Id": data["Id"],
                "RepoDigests": data.get("RepoDigests", []),
                "Config.Image": data["Config"]["Image"],
                "ContainerConfig.Image": data["ContainerConfig"]["Image"],
                "Architecture": data["Architecture"],
            }
            log.debug("scanned image info: {}:{}\n{}", image, digest, json.dumps(summary, indent=2))
            already_exists = 0
            config_digest = data["Id"]
            async with self.db.begin_readonly_session() as db_session:
                already_exists = await db_session.scalar(
                    sa.select([sa.func.count(ImageRow.id)]).where(
                        ImageRow.config_digest == config_digest,
                        ImageRow.is_local == sa.false(),
                    )
                )
            if already_exists > 0:
                return {}, "already synchronized from a remote registry"
            return {
                architecture: {
                    "size": data["Size"],
                    "labels": data["Config"]["Labels"],
                    "digest": config_digest,
                },
            }, None

        async with concurrency_sema.get():
            manifests, skip_reason = await _read_image_info(digest)
            await self._read_manifest(image, digest, manifests, skip_reason)
